"""Implementation of Rule AM07."""

from typing import Optional

from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.utils.analysis.query import Query, Selectable, WildcardInfo


class Rule_AM07(BaseRule):
    """Queries within set query produce different numbers of columns.

    **Anti-pattern**

    When writing set expressions, all queries must return the same number of columns.

    .. code-block:: sql

        WITH cte AS (
            SELECT
                a,
                b
            FROM foo
        )
        SELECT * FROM cte
        UNION
        SELECT
            c,
            d,
            e
         FROM t

    **Best practice**

    Always specify columns when writing set queries
    and ensure that they all seleect same number of columns

    .. code-block:: sql

        WITH cte AS (
            SELECT a, b FROM foo
        )
        SELECT
            a,
            b
        FROM cte
        UNION
        SELECT
            c,
            d
        FROM t
    """

    name = "ambiguous.set_columns"
    aliases = ("L068",)
    groups: tuple[str, ...] = ("all", "ambiguous")
    crawl_behaviour = SegmentSeekerCrawler({"set_expression"}, provide_raw_stack=True)

    def __resolve_wild_query(
        self,
        query: Query,
    ) -> tuple[int, bool]:
        """Attempt to resolve a full query which may contain wildcards.

        NOTE: This requires a ``Query`` as input rather than just a
        ``Selectable`` and will delegate to ``__resolve_selectable``
        once any Selectables have been identified.

        This method is *not* called on the initial set expression as
        that is evaluated as a series of Selectables. This method is
        only called on any subqueries (which may themselves be SELECT,
        WITH or set expressions) found during the resolution of any
        wildcards.
        """
        self.logger.debug("Resolving query of type %s", query.query_type)
        for s in query.selectables:
            self.logger.debug("   ...with selectable %r", s.selectable.raw)

        # if one of the source queries for a query within the set is a
        # set expression, just use the first query. If that first query isn't
        # reflective of the others, that will be caught when that segment
        # is processed. We'll know if we're in a set based on whether there
        # is more than one selectable. i.e. Just take the first selectable.
        return self.__resolve_selectable(query.selectables[0], query)

    def __resolve_selectable_wildcard(
        self, wildcard: WildcardInfo, selectable: Selectable, root_query: Query
    ) -> tuple[int, bool]:
        """Attempt to resolve a single wildcard (*) within a Selectable.

        NOTE: This means resolving the number of columns implied by
        a single *. This method would be run multiple times if there
        are multiple wildcards in a single selectable.
        """
        resolved = True
        # If there is no table specified, it is likely a subquery.
        # Handle that first.
        if not wildcard.tables:
            # Crawl the Query looking for the subquery, probably in the FROM.
            for o in root_query.crawl_sources(selectable.selectable):
                if isinstance(o, Query):
                    return self.__resolve_wild_query(o)
            # We should find one. This is not an expected path to be in.
            return 0, False  # pragma: no cover

        # There might be multiple tables referenced in some wildcard cases.
        num_cols = 0
        for wildcard_table in wildcard.tables:
            cte_name = wildcard_table
            # Get the AliasInfo for the table referenced in the wildcard
            # expression.
            alias_info = selectable.find_alias(wildcard_table)
            # attempt to resolve alias or table name to a cte
            if alias_info:
                # Crawl inside the FROM expression looking for something to
                # resolve to.
                select_info_target = next(
                    root_query.crawl_sources(alias_info.from_expression_element)
                )

                if isinstance(select_info_target, str):
                    cte_name = select_info_target
                else:
                    _cols, _resolved = self.__resolve_wild_query(select_info_target)
                    num_cols += _cols
                    resolved = resolved and _resolved
                    continue

            cte = root_query.lookup_cte(cte_name)
            if cte:
                _cols, _resolved = self.__resolve_wild_query(cte)
                num_cols += _cols
                resolved = resolved and _resolved
            else:
                # Unable to resolve
                resolved = False
        return num_cols, resolved

    def __resolve_selectable(
        self, selectable: Selectable, root_query: Query
    ) -> tuple[int, bool]:
        """Resolve the number of columns in a single Selectable.

        The selectable may or may not have wildcard (*) expressions.
        If it does, we attempt to resolve them.
        """
        self.logger.debug("Resolving selectable: %r", selectable.selectable.raw)
        # Handle cases where select_info is None (e.g., DuckDB FROM-first syntax)
        # When SELECT is omitted, it's treated as an implicit SELECT *
        if not selectable.select_info:
            self.logger.debug(
                "No select_info (e.g., implicit SELECT *): %r",
                selectable.selectable.raw,
            )
            # Cannot resolve implicit wildcards, return unresolved
            return 0, False
        wildcard_info = selectable.get_wildcard_info()
        # Start with the number of non-wild columns.
        num_cols = len(selectable.select_info.select_targets) - len(wildcard_info)

        # If there's no wildcard, just count the columns and move on.
        if not wildcard_info:
            # if there is no wildcard in the query use the count of select targets
            self.logger.debug("Resolved N=%s: %r", num_cols, selectable.selectable.raw)
            return num_cols, True

        resolved = True
        # If the set query contains on or more wildcards, attempt to resolve it to a
        # list of select targets that can be counted.
        for wildcard in wildcard_info:
            _cols, _resolved = self.__resolve_selectable_wildcard(
                wildcard, selectable, root_query
            )
            resolved = resolved and _resolved
            # Add on the number of columns which the wildcard resolves to.
            num_cols += _cols

        self.logger.debug(
            "%s N=%s: %r",
            "Resolved" if resolved else "Unresolved",
            num_cols,
            selectable.selectable.raw,
        )
        return num_cols, resolved

    def _get_select_target_counts(self, query: Query) -> tuple[set[int], bool]:
        """Given a set expression, get the number of select targets in each query.

        We keep track of the number of columns in each selectable using a
        ``set``. Ideally at the end there is only one item in the set,
        showing that all selectables have the same size. Importantly we
        can't guarantee that we can always resolve any wildcards (*), so
        we also return a flag to indicate whether any present have been
        fully resolved.
        """
        select_target_counts = set()
        resolved_wildcard = True

        for selectable in query.selectables:
            cnt, res = self.__resolve_selectable(selectable, query)
            if not res:
                resolved_wildcard = False
            select_target_counts.add(cnt)

        return select_target_counts, resolved_wildcard

    def _eval(self, context: RuleContext) -> Optional[LintResult]:
        """All queries in set expression should return the same number of columns."""
        assert context.segment.is_type("set_expression")
        root = context.segment

        # Is the parent of the set expression a WITH expression?
        # NOTE: Backward slice to work outward.
        for parent in context.parent_stack[::-1]:
            if parent.is_type("with_compound_statement"):
                # If it is, work from there instead.
                root = parent
                break

        query: Query = Query.from_segment(root, dialect=context.dialect)
        set_segment_select_sizes, resolve_wildcard = self._get_select_target_counts(
            query
        )
        self.logger.info(
            "Resolved select sizes (resolved wildcard: %s) : %s",
            resolve_wildcard,
            set_segment_select_sizes,
        )
        # if queries had different select target counts
        # and all wildcards have been resolved; fail
        if len(set_segment_select_sizes) > 1 and resolve_wildcard:
            return LintResult(anchor=context.segment)

        return LintResult()



"""Helpers for handling slices."""


def to_tuple(s: slice) -> tuple[int, int]:
    """Convert a slice into a tuple of (start, stop)."""
    assert s.start is not None and s.stop is not None
    return (s.start, s.stop)


def slice_length(s: slice) -> int:
    """Get the length of a slice."""
    length: int = s.stop - s.start
    return length


def is_zero_slice(s: slice) -> bool:
    """Return true if this is a zero slice."""
    is_zero: bool = s.stop == s.start
    return is_zero


def zero_slice(i: int) -> slice:
    """Construct a zero slice from a single integer."""
    return slice(i, i)


def offset_slice(start: int, offset: int) -> slice:
    """Construct a slice from a start and offset."""
    return slice(start, start + offset)


def slice_overlaps(s1: slice, s2: slice) -> bool:
    """Check whether two slices overlap.

    NOTE: This is designed only for use with *closed* and
    *positive* slices.
    """
    assert s1.start is not None, f"{s1} is not closed"
    assert s1.stop is not None, f"{s1} is not closed"
    assert s2.start is not None, f"{s2} is not closed"
    assert s2.stop is not None, f"{s2} is not closed"
    assert s1.start <= s1.stop, f"{s1} is not positive"
    assert s2.start <= s2.stop, f"{s2} is not positive"

    if s2.start >= s1.stop:
        return False

    if s1.start >= s2.stop:
        return False

    return True



"""Implementation of Rule CP05."""

from sqlfluff.core.parser import BaseSegment
from sqlfluff.core.rules.base import LintResult
from sqlfluff.core.rules.context import RuleContext
from sqlfluff.core.rules.crawlers import SegmentSeekerCrawler
from sqlfluff.rules.capitalisation.CP01 import Rule_CP01


class Rule_CP05(Rule_CP01):
    """Inconsistent capitalisation of datatypes.

    **Anti-pattern**

    In this example, ``int`` and ``unsigned`` are in lower-case whereas
    ``VARCHAR`` is in upper-case.

    .. code-block:: sql

        CREATE TABLE t (
            a int unsigned,
            b VARCHAR(15)
        );

    **Best practice**

    Ensure all datatypes are consistently upper or lower case

    .. code-block:: sql

        CREATE TABLE t (
            a INT UNSIGNED,
            b VARCHAR(15)
        );

    """

    name = "capitalisation.types"
    aliases = ("L063",)
    groups = ("all", "core", "capitalisation")
    is_fix_compatible = True

    crawl_behaviour = SegmentSeekerCrawler(
        {
            "data_type_identifier",
            "primitive_type",
            "datetime_type_identifier",
            "data_type",
        }
    )
    # NOTE: CP05 overrides `_eval` and then only calls
    # `_handle_segment` from CP01. Setting `_exclude_types`
    # and `_exclude_parent_types` therefore has no effect.
    # They are set here to empty tuples to avoid confusion.
    _exclude_types = ()
    _exclude_parent_types = ()
    config_keywords = [
        "extended_capitalisation_policy",
        "ignore_words",
        "ignore_words_regex",
    ]
    _description_elem = "Datatypes"

    def _eval(self, context: RuleContext) -> list[LintResult]:
        """Inconsistent capitalisation of datatypes.

        We use the `memory` feature here to keep track of cases known to be
        inconsistent with what we've seen so far as well as the top choice
        for what the possible case is.

        """
        results = []
        # For some of these segments we want to run the code on
        if context.segment.is_type(
            "primitive_type", "datetime_type_identifier", "data_type"
        ):
            for seg in context.segment.segments:
                # We don't want to edit symbols, quoted things or identifiers
                # if they appear.
                if seg.is_type(
                    "symbol", "identifier", "quoted_literal"
                ) or not seg.is_type("raw"):
                    continue
                res = self._handle_segment(seg, context)
                if res:
                    results.append(res)

        # NOTE: Given the dialect structure we can assume the targets have a parent.
        parent: BaseSegment = context.parent_stack[-1]
        # Don't process it if it's likely to have been processed by the parent.
        if context.segment.is_type("data_type_identifier") and not parent.is_type(
            "primitive_type", "datetime_type_identifier", "data_type"
        ):
            results.append(
                self._handle_segment(context.segment, context)
            )  # pragma: no cover

        return results



"""Helper utilities for reflow."""

import logging
from collections.abc import Iterable
from itertools import chain

from sqlfluff.core.parser import BaseSegment, RawSegment
from sqlfluff.core.rules import LintFix, LintResult

# We're in the utils module, but users will expect reflow
# logs to appear in the context of rules. Hence it's a subset
# of the rules logger.
reflow_logger = logging.getLogger("sqlfluff.rules.reflow")


def fixes_from_results(results: Iterable[LintResult]) -> list[LintFix]:
    """Return a list of fixes from an iterable of LintResult."""
    return list(chain.from_iterable(result.fixes for result in results))


def pretty_segment_name(segment: BaseSegment) -> str:
    """Get a nicely formatted name of the segment."""
    if segment.is_type("symbol"):
        # In a symbol reference, show the raw value and type.
        # (With underscores as spaces)
        return segment.get_type().replace("_", " ") + f" {segment.raw!r}"
    elif segment.is_type("keyword"):
        # Reference keywords as keywords.
        return f"{segment.raw!r} keyword"
    else:
        # Reference other segments just by their type.
        # (With underscores as spaces)
        return segment.get_type().replace("_", " ")


def deduce_line_indent(raw_segment: RawSegment, root_segment: BaseSegment) -> str:
    """Given a raw segment, deduce the indent of its line."""
    seg_idx = root_segment.raw_segments.index(raw_segment)
    indent_seg = None
    # Use range and a lookup here because it's more efficient than slicing
    # as we only need a subset of the long series.
    for idx in range(seg_idx, -1, -1):
        seg = root_segment.raw_segments[idx]
        if seg.is_code:
            indent_seg = None
        elif seg.is_type("whitespace"):
            indent_seg = seg
        elif seg.is_type("newline"):
            break
    reflow_logger.debug("Deduced indent for %s as %s", raw_segment, indent_seg)
    return indent_seg.raw if indent_seg else ""



"""AnyNumberOf, OneOf, OptionallyBracketed & AnySetOf."""

from collections.abc import Sequence as SequenceType
from typing import Optional, Union, cast

from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.grammar.base import (
    BaseGrammar,
    cached_method_for_parse_context,
)
from sqlfluff.core.parser.grammar.sequence import Bracketed, Sequence
from sqlfluff.core.parser.match_algorithms import (
    longest_match,
    skip_start_index_forward_to_code,
    trim_to_terminator,
)
from sqlfluff.core.parser.match_result import MatchResult
from sqlfluff.core.parser.matchable import Matchable
from sqlfluff.core.parser.segments import BaseSegment, UnparsableSegment
from sqlfluff.core.parser.types import ParseMode, SimpleHintType


def _parse_mode_match_result(
    segments: SequenceType[BaseSegment],
    current_match: MatchResult,
    max_idx: int,
    parse_mode: ParseMode,
) -> MatchResult:
    """A helper function for the return values of AnyNumberOf.

    This method creates UnparsableSegments as appropriate
    depending on the parse mode and return values.
    """
    # If we're being strict, just return.
    if parse_mode == ParseMode.STRICT:
        return current_match

    # Nothing in unmatched anyway?
    _stop_idx = current_match.matched_slice.stop
    if _stop_idx == max_idx or all(not s.is_code for s in segments[_stop_idx:max_idx]):
        return current_match

    _trim_idx = skip_start_index_forward_to_code(segments, _stop_idx)

    # Create an unmatched segment
    _expected = "Nothing else"
    if len(segments) > max_idx:
        _expected += f" before {segments[max_idx].raw!r}"

    unmatched_match = MatchResult(
        matched_slice=slice(_trim_idx, max_idx),
        matched_class=UnparsableSegment,
        segment_kwargs={"expected": _expected},
    )

    return current_match.append(unmatched_match)


class AnyNumberOf(BaseGrammar):
    """A more configurable version of OneOf."""

    supported_parse_modes = {
        ParseMode.STRICT,
        ParseMode.GREEDY,
    }

    def __init__(
        self,
        *args: Union[Matchable, str],
        max_times: Optional[int] = None,
        min_times: int = 0,
        max_times_per_element: Optional[int] = None,
        exclude: Optional[Matchable] = None,
        terminators: SequenceType[Union[Matchable, str]] = (),
        reset_terminators: bool = False,
        allow_gaps: bool = True,
        optional: bool = False,
        parse_mode: ParseMode = ParseMode.STRICT,
    ) -> None:
        self.max_times = max_times
        self.min_times = min_times
        self.max_times_per_element = max_times_per_element
        # Any patterns to _prevent_ a match.
        self.exclude = exclude
        super().__init__(
            *args,
            allow_gaps=allow_gaps,
            optional=optional,
            terminators=terminators,
            reset_terminators=reset_terminators,
            parse_mode=parse_mode,
        )

    @cached_method_for_parse_context
    def simple(
        self, parse_context: ParseContext, crumbs: Optional[tuple[str]] = None
    ) -> SimpleHintType:
        """Does this matcher support a uppercase hash matching route?

        AnyNumberOf does provide this, as long as *all* the elements *also* do.
        """
        option_simples: list[SimpleHintType] = [
            opt.simple(parse_context=parse_context, crumbs=crumbs)
            for opt in self._elements
        ]
        if any(elem is None for elem in option_simples):
            return None
        # We now know that there are no Nones.
        simple_buff = cast(list[tuple[frozenset[str], frozenset[str]]], option_simples)
        # Combine the lists
        simple_raws = [simple[0] for simple in simple_buff if simple[0]]
        simple_types = [simple[1] for simple in simple_buff if simple[1]]
        return (
            frozenset.union(*simple_raws) if simple_raws else frozenset(),
            frozenset.union(*simple_types) if simple_types else frozenset(),
        )

    def is_optional(self) -> bool:
        """Return whether this element is optional.

        This is mostly set in the init method, but also in this
        case, if min_times is zero then this is also optional.
        """
        return self.optional or self.min_times == 0

    def match(
        self,
        segments: SequenceType["BaseSegment"],
        idx: int,
        parse_context: "ParseContext",
    ) -> MatchResult:
        """Match against any of the elements a relevant number of times.

        If it matches multiple, it returns the longest, and if any are the same
        length it returns the first (unless we explicitly just match first).
        """
        if self.exclude:
            with parse_context.deeper_match(
                name=self.__class__.__name__ + "-Exclude"
            ) as ctx:
                if self.exclude.match(segments, idx, ctx):
                    return MatchResult.empty_at(idx)

        n_matches = 0
        # Keep track of the number of times each option has been matched.
        option_counter = {elem.cache_key(): 0 for elem in self._elements}
        # Keep track of how far we've got.
        matched_idx = idx
        # The working index is to cover non-code elements which aren't
        # claimed yet, but we should conditionally claim if the next
        # match is successful.
        working_idx = idx
        matched = MatchResult.empty_at(idx)
        max_idx = len(segments)  # What is the limit

        if self.parse_mode == ParseMode.GREEDY:
            max_idx = trim_to_terminator(
                segments,
                idx,
                terminators=(
                    # Only pass through the context terminators if not resetting.
                    self.terminators
                    if self.reset_terminators
                    else [*self.terminators, *parse_context.terminators]
                ),
                parse_context=parse_context,
            )

        while True:
            if n_matches >= self.min_times:
                if (
                    # Either nothing left to match...
                    matched_idx >= max_idx
                    # ...Or we've matched as many times as allowed.
                    or (self.max_times and n_matches >= self.max_times)
                ):
                    # NOTE: For OneOf, this is the matched return path.
                    return _parse_mode_match_result(
                        segments,
                        matched,
                        max_idx,
                        self.parse_mode,
                    )

            # Is there nothing left to match?
            if matched_idx >= max_idx:
                # Return unsuccessful as we didn't meet the hurdle.
                # The positive exhausted return is above.
                return MatchResult.empty_at(idx)

            with parse_context.deeper_match(
                name=self.__class__.__name__,
                clear_terminators=self.reset_terminators,
                push_terminators=self.terminators,
            ) as ctx:
                match, matched_option = longest_match(
                    # TODO: Resolve re-slice limit hack
                    segments[:max_idx],
                    self._elements,
                    working_idx,
                    ctx,
                )

            # Did we fail to match?
            if not match:
                # If we haven't already met the hurdle rate, act as though
                # not match at all.
                if n_matches < self.min_times:
                    matched = MatchResult.empty_at(idx)

                return _parse_mode_match_result(
                    segments,
                    matched,
                    max_idx,
                    self.parse_mode,
                )

            # Otherwise we have a new clean match.
            assert match
            assert matched_option

            # Update counts of each option in case we've hit limits.
            matched_key = matched_option.cache_key()
            if matched_option.cache_key() in option_counter:
                option_counter[matched_key] += 1
                # Check if we have matched an option too many times.
                if (
                    self.max_times_per_element
                    and option_counter[matched_key] > self.max_times_per_element
                ):
                    # Return the match so far, without the most recent match.
                    return _parse_mode_match_result(
                        segments,
                        matched,
                        max_idx,
                        self.parse_mode,
                    )

            # If we haven't hit limits then consume and move on.
            matched = matched.append(match)
            matched_idx = matched.matched_slice.stop
            working_idx = matched_idx
            if self.allow_gaps:
                working_idx = skip_start_index_forward_to_code(segments, matched_idx)
            parse_context.update_progress(matched_idx)
            n_matches += 1
            # Continue around the loop...


class OneOf(AnyNumberOf):
    """Match any of the elements given once.

    If it matches multiple, it returns the longest, and if any are the same
    length it returns the first (unless we explicitly just match first).
    """

    def __init__(
        self,
        *args: Union[Matchable, str],
        exclude: Optional[Matchable] = None,
        terminators: SequenceType[Union[Matchable, str]] = (),
        reset_terminators: bool = False,
        allow_gaps: bool = True,
        optional: bool = False,
        parse_mode: ParseMode = ParseMode.STRICT,
    ) -> None:
        super().__init__(
            *args,
            max_times=1,
            min_times=1,
            exclude=exclude,
            terminators=terminators,
            reset_terminators=reset_terminators,
            allow_gaps=allow_gaps,
            optional=optional,
            parse_mode=parse_mode,
        )


class OptionallyBracketed(OneOf):
    """Hybrid of Bracketed and Sequence: allows brackets but they aren't required.

    NOTE: This class is greedy on brackets so if they *can* be claimed, then
    they will be.
    """

    def __init__(
        self,
        *args: Union[Matchable, str],
        exclude: Optional[Matchable] = None,
        terminators: SequenceType[Union[Matchable, str]] = (),
        reset_terminators: bool = False,
        optional: bool = False,
        parse_mode: ParseMode = ParseMode.STRICT,
    ) -> None:
        super().__init__(
            Bracketed(*args),
            # In the case that there is only one argument, no sequence is required.
            args[0] if len(args) == 1 else Sequence(*args),
            exclude=exclude,
            terminators=terminators,
            reset_terminators=reset_terminators,
            optional=optional,
            parse_mode=parse_mode,
        )


class AnySetOf(AnyNumberOf):
    """Match any number of the elements but each element can only be matched once."""

    def __init__(
        self,
        *args: Union[Matchable, str],
        max_times: Optional[int] = None,
        min_times: int = 0,
        exclude: Optional[Matchable] = None,
        terminators: SequenceType[Union[Matchable, str]] = (),
        reset_terminators: bool = False,
        allow_gaps: bool = True,
        optional: bool = False,
        parse_mode: ParseMode = ParseMode.STRICT,
    ) -> None:
        super().__init__(
            *args,
            max_times_per_element=1,
            max_times=max_times,
            min_times=min_times,
            exclude=exclude,
            terminators=terminators,
            reset_terminators=reset_terminators,
            allow_gaps=allow_gaps,
            optional=optional,
            parse_mode=parse_mode,
        )



"""Base implementation for the plugin."""

from typing import Any

from sqlfluff.core.config import load_config_resource
from sqlfluff.core.plugin import hookimpl
from sqlfluff.core.rules import BaseRule, ConfigInfo
from sqlfluff.core.rules.config_info import STANDARD_CONFIG_INFO_DICT
from sqlfluff.core.rules.loader import get_rules_from_path
from sqlfluff.core.templaters import RawTemplater, core_templaters


@hookimpl
def get_rules() -> list[type[BaseRule]]:
    """Get plugin rules.

    NOTE: All standard rules will eventually be loaded as
    plugins and so before 2.0.0, once all legacy plugin definitions
    are migrated, this function will be amended to return no rules.
    """
    return get_rules_from_path()


@hookimpl
def get_templaters() -> list[type[RawTemplater]]:
    """Get templaters."""
    templaters = list(t for t in core_templaters())
    return templaters


@hookimpl
def load_default_config() -> dict[str, Any]:
    """Loads the default configuration for the plugin."""
    return load_config_resource(
        package="sqlfluff.core",
        file_name="default_config.cfg",
    )


@hookimpl
def get_configs_info() -> dict[str, ConfigInfo]:
    """Get rule config validations and descriptions."""
    return STANDARD_CONFIG_INFO_DICT



"""Implementation of Rule LT01."""

from typing import Optional

from sqlfluff.core.rules import BaseRule, LintResult, RuleContext
from sqlfluff.core.rules.crawlers import RootOnlyCrawler
from sqlfluff.utils.reflow.sequence import ReflowSequence


class Rule_LT01(BaseRule):
    """Inappropriate Spacing.

    This rule checks for an enforces the spacing as configured in
    :ref:`layoutconfig`. This includes excessive whitespace,
    trailing whitespace at the end of a line and also the wrong
    spacing between elements on the line. Because of this wide reach
    you may find that you wish to add specific configuration in your
    project to tweak how specific elements are treated. Rather than
    configuration on this specific rule, use the `sqlfluff.layout`
    section of your configuration file to customise how this rule
    operates.

    The ``•`` character represents a space in the examples below.

    **Anti-pattern**

    .. code-block:: sql
        :force:

        SELECT
            a,        b(c) as d••
        FROM foo••••
        JOIN bar USING(a)

    **Best practice**

    * Unless an indent or preceding a comment, whitespace should
      be a single space.

    * There should also be no trailing whitespace at the ends of lines.

    * There should be a space after :code:`USING` so that it's not confused
      for a function.

    .. code-block:: sql

        SELECT
            a, b(c) as d
        FROM foo
        JOIN bar USING (a)
    """

    name = "layout.spacing"
    # NOTE: This rule combines the following legacy rules:
    # - L001: Trailing Whitespace
    # - L005 & L008: Space around commas
    # - L006: Space around operators
    # - L023: Space after AS in WITH clause
    # - L024: Space immediately after USING
    # - L039: Unnecessary Whitespace
    # - L048: Spacing around quoted literals
    # - L071: Spacing around brackets
    aliases = ("L001", "L005", "L006", "L008", "L023", "L024", "L039", "L048", "L071")
    groups = ("all", "core", "layout")
    crawl_behaviour = RootOnlyCrawler()
    is_fix_compatible = True

    def _eval(self, context: RuleContext) -> Optional[list[LintResult]]:
        """Unnecessary whitespace."""
        sequence = ReflowSequence.from_root(context.segment, config=context.config)
        return sequence.respace().get_results()



"""Helper utilities for identifiers.

These are primarily common functions used by multiple rule
bundles. Defined here to avoid duplication, but also avoid
circular imports.
"""

from sqlfluff.core.parser import BaseSegment


def identifiers_policy_applicable(
    policy: str, parent_stack: tuple[BaseSegment, ...]
) -> bool:
    """Does `(un)quoted_identifiers_policy` apply to this segment?

    This method is used in CP02, RF04 and RF05.
    """
    if policy == "all":
        return True
    if policy == "none":
        return False
    is_alias = parent_stack and parent_stack[-1].is_type(
        "alias_expression", "column_definition", "with_compound_statement"
    )
    if policy == "aliases" and is_alias:
        return True
    is_inside_from = any(p.is_type("from_clause") for p in parent_stack)
    if policy == "column_aliases" and is_alias and not is_inside_from:
        return True
    if policy == "table_aliases" and is_alias and is_inside_from:
        return True
    return False



"""The layout plugin bundle."""

from sqlfluff.core.plugin import hookimpl
from sqlfluff.core.rules import BaseRule, ConfigInfo


@hookimpl
def get_configs_info() -> dict[str, ConfigInfo]:
    """Get additional rule config validations and descriptions."""
    return {
        "ignore_comment_lines": {
            "validation": [True, False],
            "definition": (
                "Should lines that contain only whitespace and comments"
                " be ignored when linting line lengths?"
            ),
        },
        "ignore_comment_clauses": {
            "validation": [True, False],
            "definition": (
                "Should comment clauses (e.g. column comments) be ignored"
                " when linting line lengths?"
            ),
        },
        "maximum_empty_lines_between_statements": {
            "validation": range(1000),
            "definition": (
                "The maximum number of empty lines allowed between statements. "
                "Note that currently, the gap _before_ and _after_ the semicolon "
                "is considered 'between' statements."
            ),
        },
        "maximum_empty_lines_inside_statements": {
            "validation": range(1000),
            "definition": (
                "The maximum number of empty lines allowed inside statements."
            ),
        },
        "maximum_empty_lines_between_batches": {
            "validation": range(1000),
            "definition": (
                "The maximum number of empty lines allowed between batches "
                "in T-SQL. Batches are separated by GO statements."
            ),
        },
        "wildcard_policy": {
            "validation": ["single", "multiple"],
            "definition": "Treatment of wildcards. Defaults to ``single``.",
        },
        "single_target_policy": {
            "validation": ["same_line", "new_line"],
            "definition": (
                "Treatment of single select targets. ``same_line`` (default) "
                "allows a single select target on the same line as ``SELECT``. "
                "``new_line`` requires even single targets to be on a new line, "
                "giving consistent formatting regardless of the number of "
                "select targets."
            ),
        },
    }


@hookimpl
def get_rules() -> list[type[BaseRule]]:
    """Get plugin rules.

    NOTE: Rules are imported only on fetch to manage import times
    when rules aren't used.
    """
    from sqlfluff.rules.layout.LT01 import Rule_LT01
    from sqlfluff.rules.layout.LT02 import Rule_LT02
    from sqlfluff.rules.layout.LT03 import Rule_LT03
    from sqlfluff.rules.layout.LT04 import Rule_LT04
    from sqlfluff.rules.layout.LT05 import Rule_LT05
    from sqlfluff.rules.layout.LT06 import Rule_LT06
    from sqlfluff.rules.layout.LT07 import Rule_LT07
    from sqlfluff.rules.layout.LT08 import Rule_LT08
    from sqlfluff.rules.layout.LT09 import Rule_LT09
    from sqlfluff.rules.layout.LT10 import Rule_LT10
    from sqlfluff.rules.layout.LT11 import Rule_LT11
    from sqlfluff.rules.layout.LT12 import Rule_LT12
    from sqlfluff.rules.layout.LT13 import Rule_LT13
    from sqlfluff.rules.layout.LT14 import Rule_LT14
    from sqlfluff.rules.layout.LT15 import Rule_LT15

    return [
        Rule_LT01,
        Rule_LT02,
        Rule_LT03,
        Rule_LT04,
        Rule_LT05,
        Rule_LT06,
        Rule_LT07,
        Rule_LT08,
        Rule_LT09,
        Rule_LT10,
        Rule_LT11,
        Rule_LT12,
        Rule_LT13,
        Rule_LT14,
        Rule_LT15,
    ]



"""Lower level routines for config file loading and caching.

Functions in this module load config from *individual* files and
resources. While some are cached, they are cached on the basis of
not processing individual files more than once.

For the cached functions it is VERY recommended to make sure they
are copied before any edits happen to them, as those edits may
propagate back up into the cache. Typically the results are passed
to `nested_combine` either immediately, or eventually after returning
which should negate this effect.
"""

import glob
import os.path
from functools import cache
from typing import Optional

from sqlfluff.core.config.ini import load_ini_string
from sqlfluff.core.config.toml import load_toml_file_config
from sqlfluff.core.config.validate import validate_config_dict
from sqlfluff.core.helpers.string import (
    split_comma_separated_string,
)
from sqlfluff.core.types import ConfigMappingType

COMMA_SEPARATED_PATH_KEYS = (
    "load_macros_from_path",
    "loader_search_path",
    "exclude_macros_from_path",
)
RESOLVE_PATH_SUFFIXES = ("_path", "_dir")
# Values in this section are arbitrary user data for templating, so the path
# heuristics below must not be applied to them.
OPAQUE_SECTION_KEY_PATH = ("templater", "jinja", "context")


def _load_raw_file_as_dict(filepath: str) -> ConfigMappingType:
    """Loads the raw dict object from file without interpolation."""
    filename = os.path.basename(filepath)
    if filename == "pyproject.toml":
        return load_toml_file_config(filepath)
    # If it's not a pyproject file, assume that it's an ini file.
    with open(filepath, mode="r") as file:
        return load_ini_string(file.read())


def _resolve_path(filepath: str, val: str) -> list[str]:
    """Resolves a path relative to the directory of filepath.

    Handles both exact paths and wildcard patterns.
    """
    directory = os.path.dirname(filepath)
    search_path = os.path.join(directory, val)

    # Use glob which handles both wildcard patterns and exact matches
    matches = glob.glob(search_path)

    # Return the matches if any were found
    return matches


def _resolve_paths_in_config(
    config: ConfigMappingType,
    filepath: str,
    logging_reference: Optional[str] = None,
    key_path: tuple[str, ...] = (),
) -> None:
    """Attempt to resolve any paths found in the config file.

    NOTE: This method is recursive to crawl the whole config object,
    and also mutates the underlying config object rather than returning it.
    """
    log_filename: str = logging_reference or filepath
    for key, val in config.items():
        # If it's a dict, recurse.
        if isinstance(val, dict):
            if key_path + (key.lower(),) != OPAQUE_SECTION_KEY_PATH:
                _resolve_paths_in_config(
                    val,
                    filepath,
                    logging_reference=logging_reference,
                    key_path=key_path + (key.lower(),),
                )
        # If it's a potential multi-path, split, resolve and join
        if key.lower() in COMMA_SEPARATED_PATH_KEYS:
            assert isinstance(val, str), (
                f"Value for {key} in {log_filename} must be a string not {type(val)}."
            )
            paths = split_comma_separated_string(val)
            resolved_paths = [
                resolved_path
                for path in paths
                for resolved_path in _resolve_path(filepath, path)
            ]
            # If no paths resolved, keep the original patterns
            config[key] = ",".join(resolved_paths) if resolved_paths else val
        # It it's a single path key, resolve it.
        elif key.lower().endswith(RESOLVE_PATH_SUFFIXES):
            assert isinstance(val, str), (
                f"Value for {key} in {log_filename} must be a string not {type(val)}."
            )
            resolved_paths = _resolve_path(filepath, val)
            config[key] = resolved_paths[0] if resolved_paths else val


@cache
def load_config_file_as_dict(filepath: str) -> ConfigMappingType:
    """Load the given config file into a dict and validate.

    This method is cached to mitigate being called multiple times.

    This doesn't manage the combination of config files within a nested
    structure, that happens further up the stack.
    """
    raw_config = _load_raw_file_as_dict(filepath)

    # The raw loaded files have some path interpolation which is necessary.
    _resolve_paths_in_config(raw_config, filepath)
    # Validate
    validate_config_dict(raw_config, filepath)

    # Return dict object (which will be cached)
    return raw_config


@cache
def load_config_string_as_dict(
    config_string: str, working_path: str, logging_reference: str
) -> ConfigMappingType:
    """Load the given config string and validate.

    This method is cached to mitigate being called multiple times.

    This doesn't manage the combination of config files within a nested
    structure, that happens further up the stack. The working path is
    necessary to resolve any paths in the config file.
    """
    raw_config = load_ini_string(config_string)

    # The raw loaded files have some path interpolation which is necessary.
    _resolve_paths_in_config(
        raw_config, working_path, logging_reference=logging_reference
    )
    # Validate
    validate_config_dict(raw_config, logging_reference)

    # Return dict object (which will be cached)
    return raw_config



"""Module for loading config."""

from __future__ import annotations

import logging
from collections.abc import Iterable
from copy import copy, deepcopy
from itertools import chain
from typing import TYPE_CHECKING, Any, Optional

import pluggy

from sqlfluff.core.config.ini import coerce_value
from sqlfluff.core.config.loader import load_config_string, load_config_up_to_path
from sqlfluff.core.config.validate import validate_config_dict
from sqlfluff.core.errors import SQLFluffUserError
from sqlfluff.core.helpers.dict import (
    dict_diff,
    iter_records_from_nested_dict,
    nested_combine,
    records_to_nested_dict,
)
from sqlfluff.core.helpers.string import (
    split_colon_separated_string,
    split_comma_separated_string,
)
from sqlfluff.core.plugin.host import get_plugin_manager
from sqlfluff.core.types import ConfigMappingType, ConfigValueOrListType

if TYPE_CHECKING:  # pragma: no cover
    from sqlfluff.core.templaters.base import RawTemplater

# Instantiate the config logger
config_logger = logging.getLogger("sqlfluff.config")


class FluffConfig:
    """The persistent object for internal methods to access configuration.

    This class is designed to be instantiated once for each file and then be
    reused by each part of the process. For multiple files in the same path, a
    parent object will be created for the each path and then variants of it
    are created *for each file*. The object itself contains the references
    to any long lived objects which might be used by multiple parts of the
    codebase such as the dialect and the templater (both of which can be
    resource intensive to load & instantiate), which allows (for example),
    multiple files to reuse the same instance of the relevant dialect.

    It is also designed to pickle well for use in parallel operations.

    Args:
        configs (ConfigMappingType, optional): A nested dict of config
            values from which to construct the config.
        extra_config_path (str, optional): An optional additional path
            to load config files from. These are loaded last if found
            and take precedence over any pre-existing config values.
            Note that when provided directly to the class, this path
            is not loaded for the class in question (it's assumed that
            has already been done, and the results are incorporated in
            the `configs` argument), but it *is* passed onward to child
            config instances, which will use it.
        ignore_local_config (bool, optional, defaults to False): If set to
            True, this skips loading configuration from the user home
            directory (``~``) or ``appdir`` path.
        overrides (ConfigMappingType, optional): A additional set of
            configs to merge into the ``core`` section of the config
            object at the end. These values take precedence over all
            other provided values and are inherited by child configs.
            For example, override values provided in the CLI use this
            method to apply to all files in a linting operation. Note
            that this mapping dict *only* applies to the ``core``
            section and so cannot be used for all values.
        plugin_manager (PluginManager, optional): Optional pre-loaded
            config manager. Generally users should not need to provide
            this, as the class will fetch it's own if not provided.
            This argument is used when creating new class instances to
            avoid reloading the manager.

    .. note::
       Methods for accessing internal properties on the config are not particularly
       standardised as the project currently assumes that few other tools are using
       this interface directly. If you or your project would like more formally
       supported methods for access to the config object, raise an issue on GitHub
       with the kind of things you'd like to achieve.
    """

    private_vals = "rule_denylist", "rule_allowlist", "dialect_obj", "templater_obj"

    def __init__(
        self,
        configs: Optional[ConfigMappingType] = None,
        extra_config_path: Optional[str] = None,
        ignore_local_config: bool = False,
        overrides: Optional[ConfigMappingType] = None,
        plugin_manager: Optional[pluggy.PluginManager] = None,
        # Ideally a dialect should be set when config is read but sometimes
        # it might only be set in nested .sqlfluff config files, so allow it
        # to be not required.
        require_dialect: bool = True,
    ) -> None:
        self._extra_config_path = (
            extra_config_path  # We only store this for child configs
        )
        self._ignore_local_config = (
            ignore_local_config  # We only store this for child configs
        )
        # If overrides are provided, validate them early.
        if overrides:
            overrides = {"core": overrides}
            validate_config_dict(overrides, "<provided overrides>")
        # Stash overrides so we can pass them to child configs
        core_overrides = overrides["core"] if overrides else None
        assert isinstance(core_overrides, dict) or core_overrides is None
        self._overrides = core_overrides

        # Fetch a fresh plugin manager if we weren't provided with one
        self._plugin_manager = plugin_manager or get_plugin_manager()

        defaults = nested_combine(*self._plugin_manager.hook.load_default_config())
        # If any existing configs are provided. Validate them:
        if configs:
            validate_config_dict(configs, "<provided configs>")
        empty_config: ConfigMappingType = {"core": {}}
        empty_overrides: ConfigMappingType = {}
        self._configs = nested_combine(
            defaults, configs or empty_config, overrides or empty_overrides
        )
        # Some configs require special treatment
        nocolor = self._configs["core"].get("nocolor", None)
        self._configs["core"]["color"] = (
            False if nocolor is True else True if nocolor is False else None
        )
        # Handle inputs which are potentially comma separated strings
        self._handle_comma_separated_values()
        # Dialect and Template selection.
        _dialect = self._configs["core"]["dialect"]
        assert _dialect is None or isinstance(_dialect, str)
        self._initialise_dialect(_dialect, require_dialect)

        self._configs["core"]["templater_obj"] = self.get_templater()

        # Guard against an impossible Rust configuration.
        self._verify_rust_config()

    def _handle_comma_separated_values(self) -> None:
        for in_key, out_key in [
            ("ignore", "ignore"),
            ("warnings", "warnings"),
            ("rules", "rule_allowlist"),
            ("exclude_rules", "rule_denylist"),
        ]:
            in_value = self._configs["core"].get(in_key, None)
            if in_value:
                assert not isinstance(in_value, dict)
                self._configs["core"][out_key] = split_comma_separated_string(in_value)
            else:
                self._configs["core"][out_key] = []

    def _initialise_dialect(
        self, dialect: Optional[str], require_dialect: bool = True
    ) -> None:
        # NB: We import here to avoid a circular references.
        from sqlfluff.core.dialects import dialect_selector

        if dialect is not None:
            self._configs["core"]["dialect_obj"] = dialect_selector(dialect)
        elif require_dialect:
            self.verify_dialect_specified()

    def verify_dialect_specified(self) -> None:
        """Check if the config specifies a dialect, raising an error if not.

        Raises:
            SQLFluffUserError: If dialect config value is unset. The content
                of the error contains user-facing instructions on what dialects
                are available and how to set the dialect.
        """
        if self._configs["core"].get("dialect", None) is None:
            # Get list of available dialects for the error message. We must
            # import here rather than at file scope in order to avoid a circular
            # import.
            from sqlfluff.core.dialects import dialect_readout

            raise SQLFluffUserError(
                "No dialect was specified. You must configure a dialect or "
                "specify one on the command line using --dialect after the "
                "command. Available dialects:\n"
                f"{', '.join([d.label for d in dialect_readout()])}"
            )

    def _verify_rust_config(self) -> None:
        """Reject a contradictory Rust parser/rules configuration.

        Rust-native rules read the parse arena, which only exists when the Rust
        parser runs. Explicitly enabling ``use_rust_rules`` while the Rust
        parser will not run (``use_rust_parser`` disabled) is contradictory —
        the rules would silently do nothing — so raise instead.

        ``auto`` degrades gracefully and is never an error: ``use_rust_rules``
        is only "on" for the explicit truthy values, and the parser is
        considered enabled for ``auto`` too (matching the linter's own logic).
        """
        core = self._configs["core"]
        # Mirror the linter's parser-enable test (linter.py).
        parser_enabled = core.get("use_rust_parser") in (
            "auto",
            "Auto",
            "AUTO",
            True,
            "True",
            "true",
            1,
        )
        rules_forced_on = core.get("use_rust_rules") in (True, "True", "true", 1)
        if rules_forced_on and not parser_enabled:
            raise SQLFluffUserError(
                "use_rust_rules is enabled but use_rust_parser is disabled. "
                "Rust-native rules require the Rust parser (they read its parse "
                "arena), so this combination does nothing. Set use_rust_parser "
                "to 'auto' or 'True', or disable use_rust_rules."
            )

    def __getstate__(self) -> dict[str, Any]:
        # Copy the object's state from self.__dict__ which contains
        # all our instance attributes. Always use the dict.copy()
        # method to avoid modifying the original state.
        state = self.__dict__.copy()
        # Remove the unpicklable entries.
        del state["_plugin_manager"]
        # The dbt templater doesn't pickle well, but isn't required
        # within threaded operations. If it was, it could easily be
        # rehydrated within the thread. For rules which want to determine
        # the type of a templater in their context, use
        # `get_templater_class()` instead, which avoids instantiating
        # a new templater instance.
        # NOTE: It's important that we do this on a copy so that we
        # don't disturb the original object if it's still in use.
        state["_configs"] = state["_configs"].copy()
        state["_configs"]["core"] = state["_configs"]["core"].copy()
        state["_configs"]["core"]["templater_obj"] = None
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:  # pragma: no cover
        # Restore instance attributes
        self.__dict__.update(state)
        # NOTE: Rather than rehydrating the previous plugin manager, we
        # fetch a fresh one.
        self._plugin_manager = get_plugin_manager()
        # NOTE: Likewise we don't reinstate the "templater_obj" config value
        # which should also only be used in the main thread rather than child
        # processes.

    def copy(self) -> FluffConfig:
        """Create a copy of this ``FluffConfig``.

        Copies created using this method can safely be modified without those
        changes propagating back up to the object which was originally copied.

        Returns:
            :obj:`FluffConfig`: A shallow copy of this config object but with
            a deep copy of the internal ``_configs`` dict.
        """
        core = self._configs["core"]
        # `dialect_obj` and `templater_obj` are large, effectively immutable
        # (never reassigned in place after construction) and safe to share
        # across copies. `dialect_obj` in particular is a fully expanded
        # grammar object graph, so deep-copying it on every `.copy()` call
        # (i.e. on every parse) would dominate parse time. Seed deepcopy's
        # memo so it reuses these references instead of duplicating them.
        shared = (core.get("dialect_obj"), core.get("templater_obj"))
        memo = {id(obj): obj for obj in shared if obj is not None}
        configs_attribute_copy = deepcopy(self._configs, memo)

        config_copy = copy(self)
        config_copy._configs = configs_attribute_copy
        return config_copy

    @classmethod
    def from_root(
        cls,
        extra_config_path: Optional[str] = None,
        ignore_local_config: bool = False,
        overrides: Optional[ConfigMappingType] = None,
        require_dialect: bool = True,
    ) -> FluffConfig:
        """Loads a config object based on the root directory.

        Args:
            extra_config_path (str, optional): An optional additional path
                to load config files from. These are loaded last if found
                and take precedence over any pre-existing config values.
            ignore_local_config (bool, optional, defaults to False): If set to
                True, this skips loading configuration from the user home
                directory (``~``) or ``appdir`` path.
            overrides (ConfigMappingType, optional): A additional set of
                configs to merge into the config object at the end. These
                values take precedence over all other provided values and
                are inherited by child configs. For example, override values
                provided in the CLI use this method to apply to all files
                in a linting operation.
            require_dialect (bool, optional, default is True): When True
                an error will be raise if the dialect config value is unset.

        Returns:
            :obj:`FluffConfig`: The loaded config object.
        """
        configs = load_config_up_to_path(
            path=".",
            extra_config_path=extra_config_path,
            ignore_local_config=ignore_local_config,
        )
        return cls(
            configs=configs,
            extra_config_path=extra_config_path,
            ignore_local_config=ignore_local_config,
            overrides=overrides,
            require_dialect=require_dialect,
        )

    @classmethod
    def from_string(
        cls,
        config_string: str,
        overrides: Optional[ConfigMappingType] = None,
    ) -> FluffConfig:
        """Loads a config object from a single config string.

        Args:
            config_string (str): The config string, assumed to be in ``ini``
                format (like a ``.sqlfluff`` file).
            overrides (ConfigMappingType, optional): A additional set of
                configs to merge into the config object at the end. These
                values take precedence over all other provided values and
                are inherited by child configs. For example, override values
                provided in the CLI use this method to apply to all files
                in a linting operation.

        Returns:
            :obj:`FluffConfig`: The loaded config object.
        """
        return cls(
            configs=load_config_string(config_string),
            overrides=overrides,
        )

    @classmethod
    def from_strings(
        cls,
        *config_strings: str,
        overrides: Optional[ConfigMappingType] = None,
    ) -> FluffConfig:
        """Loads a config object given a series of nested config strings.

        Args:
            *config_strings (str): An iterable of config strings, assumed
                to be in ``ini`` format (like a ``.sqlfluff`` file).
            overrides (ConfigMappingType, optional): A additional set of
                configs to merge into the config object at the end. These
                values take precedence over all other provided values and
                are inherited by child configs. For example, override values
                provided in the CLI use this method to apply to all files
                in a linting operation.

        Returns:
            :obj:`FluffConfig`: The loaded config object.

        Config strings are incorporated from first to last, treating the
        first element as the "root" config, and then later config strings
        will take precedence over any earlier values.
        """
        config_state: ConfigMappingType = {}
        for config_string in config_strings:
            config_state = load_config_string(config_string, configs=config_state)
        return cls(
            configs=config_state,
            overrides=overrides,
        )

    @classmethod
    def from_path(
        cls,
        path: str,
        extra_config_path: Optional[str] = None,
        ignore_local_config: bool = False,
        overrides: Optional[ConfigMappingType] = None,
        plugin_manager: Optional[pluggy.PluginManager] = None,
        require_dialect: bool = True,
    ) -> FluffConfig:
        """Loads a config object given a particular path.

        Args:
            path (str): The target path to load config files from. Files
                found between the working path and this path are also loaded
                and nested with files closest to this target path taking
                precedence.
            extra_config_path (str, optional): An optional additional path
                to load config files from. These are loaded last if found
                and take precedence over any pre-existing config values.
            ignore_local_config (bool, optional, defaults to False): If set to
                True, this skips loading configuration from the user home
                directory (``~``) or ``appdir`` path.
            overrides (ConfigMappingType, optional): A additional set of
                configs to merge into the ``core`` section of the config
                object at the end. These values take precedence over all
                other provided values and are inherited by child configs.
                Note that this mapping dict *only* applies to the ``core``
                section and so cannot be used for all values.
            plugin_manager (PluginManager, optional): Optional pre-loaded
                config manager. Generally users should not need to provide
                this, as the class will fetch it's own if not provided.
                This argument is used when creating new class instances to
                avoid reloading the manager.
            require_dialect (bool, optional, default is True): When True
                an error will be raise if the dialect config value is unset.

        Returns:
            :obj:`FluffConfig`: The loaded config object.
        """
        configs = load_config_up_to_path(
            path=path,
            extra_config_path=extra_config_path,
            ignore_local_config=ignore_local_config,
        )
        return cls(
            configs=configs,
            extra_config_path=extra_config_path,
            ignore_local_config=ignore_local_config,
            overrides=overrides,
            plugin_manager=plugin_manager,
            require_dialect=require_dialect,
        )

    @classmethod
    def from_kwargs(
        cls,
        dialect: Optional[str] = None,
        rules: Optional[list[str]] = None,
        exclude_rules: Optional[list[str]] = None,
        require_dialect: bool = True,
    ) -> FluffConfig:
        """Instantiate a config from a subset of common options.

        Args:
            dialect (str, optional): The name of the dialect to use.
            rules (list of str, optional): A list of rules to include.
                Rule specifiers can be codes, names, groups or aliases.
                If not set, defaults to all rules.
            exclude_rules (list of str, optional): A list of rules to
                exclude. Rule specifiers can be codes, names, groups or
                aliases. If not set, does not exclude any rules.
            require_dialect (bool, optional, default is True): When True
                an error will be raise if the dialect config value is unset.

        Returns:
            :obj:`FluffConfig`: The loaded config object.

        This is a convenience method for the ways that the public classes
        like Linter(), Parser() and Lexer() allow a subset of attributes to
        be set directly rather than requiring a pre-made `FluffConfig`.
        """
        overrides: ConfigMappingType = {}
        if dialect:
            overrides["dialect"] = dialect
        if rules:
            # Make a comma separated string to pass in as override
            overrides["rules"] = ",".join(rules)
        if exclude_rules:
            # Make a comma separated string to pass in as override
            overrides["exclude_rules"] = ",".join(exclude_rules)

        return cls(overrides=overrides, require_dialect=require_dialect)

    def get_templater_class(self) -> type[RawTemplater]:
        """Get the configured templater class.

        .. note::
           This is mostly useful to call directly when rules want to determine
           the *type* of a templater without (in particular to work out if it's a
           derivative of the jinja templater), without needing to instantiate a
           full templater. Instantiated templaters don't pickle well, so aren't
           automatically passed around between threads/processes.
        """
        templater_lookup: dict[str, type[RawTemplater]] = {
            templater.name: templater
            for templater in chain.from_iterable(
                self._plugin_manager.hook.get_templaters()
            )
        }
        # Fetch the config value.
        templater_name = self._configs["core"].get("templater", "<no value set>")
        assert isinstance(templater_name, str), (
            f"Config value `templater` expected to be a string. Not: {templater_name!r}"
        )
        try:
            cls = templater_lookup[templater_name]
            # Return class. Do not instantiate yet. That happens in `get_templater()`
            # for situations which require it.
            return cls
        except KeyError:
            if templater_name == "dbt":  # pragma: no cover
                config_logger.warning(
                    "Starting in sqlfluff version 0.7.0 the dbt templater is "
                    "distributed as a separate python package. Please pip install "
                    "sqlfluff-templater-dbt to use it."
                )
            raise SQLFluffUserError(
                "Requested templater {!r} which is not currently available. Try one of "
                "{}".format(templater_name, ", ".join(templater_lookup.keys()))
            )

    def get_templater(self, **kwargs: Any) -> RawTemplater:
        """Instantiate the configured templater."""
        return self.get_templater_class()(**kwargs)

    def make_child_from_path(
        self, path: str, require_dialect: bool = True
    ) -> FluffConfig:
        """Make a child config at a path but pass on overrides and extra_config_path.

        Args:
            path (str): The path to load the new config object from, inheriting
                the content of the calling `FluffConfig` as base values.
            require_dialect (bool, optional, default is True): When True
                an error will be raise if the dialect config value is unset.

        Returns:
            :obj:`FluffConfig`: A new config object which copies the current
            config object, but overriding any values set by config values loaded
            from the given path.
        """
        return self.from_path(
            path,
            extra_config_path=self._extra_config_path,
            ignore_local_config=self._ignore_local_config,
            overrides=self._overrides,
            plugin_manager=self._plugin_manager,
            require_dialect=require_dialect,
        )

    def diff_to(self, other: FluffConfig) -> ConfigMappingType:
        """Compare this config to another.

        This is primarily used in the CLI logs to indicate to the user
        what values have been changed for each file compared to the root
        config for the project.

        Args:
            other (:obj:`FluffConfig`): Another config object to compare
                against. We will return keys from *this* object that are
                not in `other` or are different to those in `other`.

        Returns:
            :obj:`dict`: A filtered dict of items in this config that are
            not in the other or are different to the other.
        """
        # We ignore some objects which are not meaningful in the comparison
        # e.g. dialect_obj, which is generated on the fly.
        return dict_diff(self._configs, other._configs, ignore=["dialect_obj"])

    def get(
        self, val: str, section: str | Iterable[str] = "core", default: Any = None
    ) -> Any:
        """Get a particular value from the config.

        Args:
            val (str): The name of the config value to get.
            section (str or iterable of str, optional): The "path" to the config
                value. For values in the main ``[sqlfluff]`` section of the
                config, which are stored in the ``core`` section of the config
                this can be omitted.
            default: The value to return if the config value was not found. If
                no default is provided, then a ``KeyError`` will be raised if
                no value was found.

        The following examples show how to fetch various default values:

        >>> FluffConfig(overrides={"dialect": "ansi"}).get("dialect")
        'ansi'

        >>> config = FluffConfig(overrides={"dialect": "ansi"})
        >>> config.get("tab_space_size", section="indentation")
        4

        >>> FluffConfig(overrides={"dialect": "ansi"}).get(
        ...     "capitalisation_policy",
        ...     section=["rules", "capitalisation.keywords"]
        ... )
        'consistent'
        """
        section_dict = self.get_section(section)
        if section_dict is None:
            return default

        return section_dict.get(val, default)

    def get_section(self, section: str | Iterable[str]) -> Any:
        """Return a whole section of config as a dict.

        If the element found at the address is a value and not
        a section, it is still returned and so this can be used
        as a more advanced from of the basic `get` method.

        Args:
            section: An iterable or string. If it's a string
                we load that root section. If it's an iterable
                of strings, then we treat it as a path within
                the dictionary structure.

        """
        if isinstance(section, str):
            return self._configs.get(section, None)
        else:
            # Try iterating
            buff = self._configs
            for sec in section:
                temp = buff.get(sec, None)
                if temp is None:
                    return None
                buff = temp
            return buff

    def set_value(self, config_path: Iterable[str], val: Any) -> None:
        """Set a value at a given path.

        Args:
            config_path: An iterable of strings. Each should be
                a one of the elements which is colon delimited in
                a standard config file.
            val: The value to set at the given path.

        >>> cfg = FluffConfig(overrides={"dialect": "ansi"})
        >>> cfg.set_value(["dialect"], "postgres")
        >>> cfg.get("dialect")
        'postgres'

        >>> cfg = FluffConfig(overrides={"dialect": "ansi"})
        >>> cfg.set_value(["indentation", "tab_space_size"], 2)
        >>> cfg.get("tab_space_size", section="indentation")
        2
        """
        # Make the path a list so we can index on it
        config_path = list(config_path)
        # Coerce the value into something more useful.
        config_val = coerce_value(val)
        # Sort out core if not there
        if len(config_path) == 1:  # pragma: no cover TODO?
            config_path = ["core"] + config_path
        # Current section:
        dict_buff = [self._configs]
        for elem in config_path[:-1]:
            dict_buff.append(dict_buff[-1].get(elem, {}))
        # Set the value
        dict_buff[-1][config_path[-1]] = config_val
        # Rebuild the config
        for elem in reversed(config_path[:-1]):
            dict_elem = dict_buff.pop()
            dict_buff[-1][elem] = dict_elem
        self._configs = dict_buff[0]

    def iter_vals(
        self, cfg: Optional[ConfigMappingType] = None
    ) -> Iterable[tuple[int, str, ConfigValueOrListType]]:
        """Return an iterable of tuples representing keys.

        Args:
            cfg (optional): An optional config mapping to format instead.
                If not provided, we use the internal config object of the
                `FluffConfig`.

        This is primarily to enable formatting of config objects in the CLI.

        We show values before dicts, the tuple contains an indent value to
        know what level of the dict we're in. Dict labels will be returned
        as a blank value before their content.
        """
        cfg = cfg or self._configs

        # Get keys and sort
        keys = sorted(cfg.keys())
        # First iterate values (alphabetically):
        for k in keys:
            value = cfg[k]
            if (
                not isinstance(value, dict)
                and value is not None
                and k not in self.private_vals
            ):
                yield (0, k, value)

        # Then iterate dicts (alphabetically (but `core` comes first if it exists))
        for k in keys:
            value = cfg[k]
            if isinstance(value, dict):
                # First yield the dict label
                yield (0, k, "")
                # Then yield its content
                for idnt, key, val in self.iter_vals(cfg=value):
                    yield (idnt + 1, key, val)

    def process_inline_config(self, config_line: str, fname: str) -> None:
        """Process an inline config command and update self.

        Args:
            config_line (str): The inline config section to be processed.
                This should usually begin with ``-- sqlfluff:``.
            fname (str): The name of the current file being processed. This
                is used purely for logging purposes in the case that an
                invalid config string is provided so that any error messages
                can reference the file with the issue.

        >>> cfg = FluffConfig(overrides={"dialect": "ansi"})
        >>> cfg.process_inline_config(
        ...     "-- sqlfluff:dialect:postgres",
        ...     "test.sql"
        ... )
        >>> cfg.get("dialect")
        'postgres'
        """
        # Strip preceding comment marks
        if config_line.startswith("--"):
            config_line = config_line[2:].strip()
        # Strip preceding sqlfluff line.
        if not config_line.startswith("sqlfluff:"):  # pragma: no cover
            config_logger.warning(
                "Unable to process inline config statement: %r", config_line
            )
            return
        config_line = config_line[9:].strip()
        config_key, config_value = split_colon_separated_string(config_line)
        # Guard against malformed inline config where the remainder after
        # stripping "sqlfluff:" contains no colon (e.g. "disable=AM04" using
        # the .sqlfluff file `=` syntax instead of the inline colon syntax).
        if len(config_key) == 0:
            config_logger.warning(
                "Malformed inline config in %r: %r. Inline config uses colon "
                "syntax (e.g. `-- sqlfluff:dialect:ansi`), not `=`.",
                fname,
                config_line,
            )
            return
        # Move to core section if appropriate
        if len(config_key) == 1:
            config_key = ("core",) + config_key
        # Coerce data types
        config_record = (config_key, coerce_value(config_value))
        # Convert to dict & validate
        config_dict: ConfigMappingType = records_to_nested_dict([config_record])
        validate_config_dict(config_dict, f"inline config in {fname}")
        config_val = list(iter_records_from_nested_dict(config_dict))[0]

        # Set the value
        self.set_value(config_key, config_value)
        # If the config is for dialect, initialise the dialect.
        if config_val[0] == ("core", "dialect"):
            dialect_value = config_val[1]
            assert isinstance(dialect_value, str)
            self._initialise_dialect(dialect_value)

    def process_raw_file_for_config(self, raw_str: str, fname: str) -> None:
        """Process a full raw file for inline config and update self.

        Args:
            raw_str (str): The full SQL script to evaluate for inline configs.
            fname (str): The name of the current file being processed. This
                is used purely for logging purposes in the case that an
                invalid config string is provided so that any error messages
                can reference the file with the issue.

        >>> cfg = FluffConfig(overrides={"dialect": "ansi"})
        >>> cfg.process_raw_file_for_config(
        ...     "-- sqlfluff:dialect:postgres",
        ...     "test.sql"
        ... )
        >>> cfg.get("dialect")
        'postgres'
        """
        # Scan the raw file for config commands.
        for raw_line in raw_str.splitlines():
            # With or without a space.
            if raw_line.startswith(("-- sqlfluff", "--sqlfluff")):
                # Found a in-file config command
                self.process_inline_config(raw_line, fname)
        # Deal with potential list-like inputs.
        self._handle_comma_separated_values()
        # Re-validate: inline config may have changed the rust parser/rules keys.
        self._verify_rust_config()



"""Defines container classes for handling noqa comments."""

import fnmatch
import logging
from dataclasses import dataclass
from typing import Any, Optional, Union, cast

from sqlfluff.core.errors import SQLBaseError, SQLParseError, SQLUnusedNoQaWarning
from sqlfluff.core.parser import BaseSegment, RawSegment, RegexLexer

# Instantiate the linter logger
linter_logger: logging.Logger = logging.getLogger("sqlfluff.linter")


@dataclass
class NoQaDirective:
    """Parsed version of a 'noqa' comment."""

    line_no: int  # Source line number
    line_pos: int  # Source line position
    rules: Optional[tuple[str, ...]]  # Affected rule names
    action: Optional[str]  # "enable", "disable", or "None"
    raw_str: str = ""  # The raw representation of the directive for warnings.
    used: bool = False  # Has it been used.

    def _filter_violations_single_line(
        self, violations: list[SQLBaseError]
    ) -> list[SQLBaseError]:
        """Filter a list of violations based on this single line noqa.

        Also record whether this class was _used_ in any of that filtering.

        The "ignore" list is assumed to ONLY contain NoQaDirectives with
        action=None.
        """
        assert not self.action
        matched_violations = [
            v
            for v in violations
            if (
                v.line_no == self.line_no
                and (self.rules is None or v.rule_code() in self.rules)
            )
        ]
        if matched_violations:
            # Successful match, mark ignore as used.
            self.used = True
            return [v for v in violations if v not in matched_violations]
        else:
            return violations


class IgnoreMask:
    """Structure to hold a set of 'noqa' directives."""

    def __init__(self, ignores: list[NoQaDirective]):
        self._ignore_list = ignores

    def __repr__(self) -> str:  # pragma: no cover
        return "<IgnoreMask>"

    # ### Construction class methods.

    @staticmethod
    def _parse_noqa(
        comment: str,
        line_no: int,
        line_pos: int,
        reference_map: dict[str, set[str]],
    ) -> Union[NoQaDirective, SQLParseError, None]:
        """Extract ignore mask entries from a comment string."""
        # Also trim any whitespace afterward

        # Comment lines can also have noqa e.g.
        # --dafhsdkfwdiruweksdkjdaffldfsdlfjksd -- noqa: LT05
        # Therefore extract last possible inline ignore.
        comment = [c.strip() for c in comment.split("--")][-1]

        if comment.startswith("noqa"):
            # This is an ignore identifier
            comment_remainder = comment[4:]
            if comment_remainder:
                if not comment_remainder.startswith(":"):
                    return SQLParseError(
                        "Malformed 'noqa' section. Expected 'noqa: <rule>[,...]",
                        line_no=line_no,
                    )
                comment_remainder = comment_remainder[1:].strip()
                if comment_remainder:
                    action: Optional[str]
                    if "=" in comment_remainder:
                        action, rule_part = comment_remainder.split("=", 1)
                        if action not in {"disable", "enable"}:  # pragma: no cover
                            return SQLParseError(
                                "Malformed 'noqa' section. "
                                "Expected 'noqa: enable=<rule>[,...] | all' "
                                "or 'noqa: disable=<rule>[,...] | all",
                                line_no=line_no,
                            )
                    else:
                        action = None
                        rule_part = comment_remainder
                        if rule_part in {"disable", "enable"}:
                            return SQLParseError(
                                "Malformed 'noqa' section. "
                                "Expected 'noqa: enable=<rule>[,...] | all' "
                                "or 'noqa: disable=<rule>[,...] | all",
                                line_no=line_no,
                            )
                    rules: Optional[tuple[str, ...]]
                    if rule_part != "all":
                        # Rules can be globs therefore we compare to the rule_set to
                        # expand the globs.
                        unexpanded_rules = tuple(
                            r.strip() for r in rule_part.split(",")
                        )
                        # We use a set to do natural deduplication.
                        expanded_rules: set[str] = set()
                        for r in unexpanded_rules:
                            matched = False
                            for expanded in (
                                reference_map[x]
                                for x in fnmatch.filter(reference_map.keys(), r)
                            ):
                                expanded_rules |= expanded
                                matched = True

                            if not matched:
                                # We were unable to expand the glob.
                                # Therefore assume the user is referencing
                                # a special error type (e.g. PRS, LXR, or TMP)
                                # and add this to the list of rules to ignore.
                                expanded_rules.add(r)
                        # Sort for consistency
                        rules = tuple(sorted(expanded_rules))
                    else:
                        rules = None
                    return NoQaDirective(line_no, line_pos, rules, action, comment)
            return NoQaDirective(line_no, line_pos, None, None, comment)
        return None

    @classmethod
    def _extract_ignore_from_comment(
        cls,
        comment: RawSegment,
        reference_map: dict[str, set[str]],
    ) -> Union[NoQaDirective, SQLParseError, None]:
        """Extract ignore mask entries from a comment segment."""
        # Also trim any whitespace
        comment_content = comment.raw_trimmed().strip()
        # If we have leading or trailing block comment markers, also strip them.
        # NOTE: We need to strip block comment markers from the start
        # to ensure that noqa directives in the following form are followed:
        # /* noqa: disable=all */
        if comment_content.endswith("*/"):
            comment_content = comment_content[:-2].rstrip()
        if comment_content.startswith("/*"):
            comment_content = comment_content[2:].lstrip()
        comment_line, comment_pos = comment.pos_marker.source_position()
        result = cls._parse_noqa(
            comment_content, comment_line, comment_pos, reference_map
        )
        if isinstance(result, SQLParseError):
            result.segment = comment
        return result

    @classmethod
    def from_tree(
        cls,
        tree: BaseSegment,
        reference_map: dict[str, set[str]],
    ) -> tuple["IgnoreMask", list[SQLBaseError]]:
        """Look for inline ignore comments and return NoQaDirectives."""
        ignore_buff: list[NoQaDirective] = []
        violations: list[SQLBaseError] = []
        for comment in tree.recursive_crawl("comment"):
            if comment.is_type("inline_comment", "block_comment"):
                ignore_entry = cls._extract_ignore_from_comment(
                    cast(RawSegment, comment), reference_map
                )
                if isinstance(ignore_entry, SQLParseError):
                    violations.append(ignore_entry)
                elif ignore_entry:
                    ignore_buff.append(ignore_entry)
        if ignore_buff:
            linter_logger.info("Parsed noqa directives from file: %r", ignore_buff)
        return cls(ignore_buff), violations

    @classmethod
    def from_source(
        cls,
        source: str,
        inline_comment_regex: RegexLexer,
        reference_map: dict[str, set[str]],
    ) -> tuple["IgnoreMask", list[SQLBaseError]]:
        """Look for inline ignore comments and return NoQaDirectives.

        Very similar to .from_tree(), but can be run on raw source
        (i.e. does not require the code to have parsed successfully).
        """
        ignore_buff: list[NoQaDirective] = []
        violations: list[SQLBaseError] = []
        for idx, line in enumerate(source.split("\n")):
            match = inline_comment_regex.search(line) if line else None
            if match:
                ignore_entry = cls._parse_noqa(
                    line[match[0] : match[1]], idx + 1, match[0], reference_map
                )
                if isinstance(ignore_entry, SQLParseError):
                    violations.append(ignore_entry)  # pragma: no cover
                elif ignore_entry:
                    ignore_buff.append(ignore_entry)
        if ignore_buff:
            linter_logger.info("Parsed noqa directives from file: %r", ignore_buff)
        return cls(ignore_buff), violations

    @classmethod
    def from_source_with_dialect(
        cls,
        source: str,
        dialect: Any,
        reference_map: dict[str, set[str]],
    ) -> tuple["IgnoreMask", list[SQLBaseError]]:
        """Parse noqa directives from raw source using a dialect's comment matcher.

        If a dialect does not define an ``inline_comment`` lexer matcher, return an
        empty ignore mask and no new violations so callers can continue without
        raising an unhelpful ``IndexError``.
        """
        inline_comment_regex = next(
            (
                cast(RegexLexer, matcher)
                for matcher in dialect.lexer_matchers
                if matcher.name == "inline_comment"
            ),
            None,
        )
        if inline_comment_regex is None:
            return cls([]), []
        return cls.from_source(source, inline_comment_regex, reference_map)

    # ### Application methods.

    @staticmethod
    def _ignore_masked_violations_single_line(
        violations: list[SQLBaseError], ignore_mask: list[NoQaDirective]
    ) -> list[SQLBaseError]:
        """Filter a list of violations based on this single line noqa.

        The "ignore" list is assumed to ONLY contain NoQaDirectives with
        action=None.
        """
        for ignore in ignore_mask:
            violations = ignore._filter_violations_single_line(violations)
        return violations

    @staticmethod
    def _should_ignore_violation_line_range(
        line_no: int, ignore_rules: list[NoQaDirective]
    ) -> tuple[bool, Optional[NoQaDirective]]:
        """Returns whether to ignore a violation at line_no.

        Loop through the NoQaDirectives to find the state of things at
        line_no. Assumptions about "ignore_rules":
        - Contains directives for only ONE RULE, i.e. the rule that was
          violated at line_no
        - Sorted in ascending order by line number
        """
        ignore = False
        last_ignore: Optional[NoQaDirective] = None
        for idx, ignore_rule in enumerate(ignore_rules):
            if ignore_rule.line_no > line_no:
                # Peak at the next rule to see if it's a matching disable
                # and if it is, then mark it as used.
                if ignore_rule.action == "enable":
                    # Mark as used
                    ignore_rule.used = True
                break

            if ignore_rule.action == "enable":
                # First, if this enable did counteract a
                # corresponding _disable_, then it has been _used_.
                if last_ignore:
                    ignore_rule.used = True
                last_ignore = None
                ignore = False
            elif ignore_rule.action == "disable":
                last_ignore = ignore_rule
                ignore = True

        return ignore, last_ignore

    @classmethod
    def _ignore_masked_violations_line_range(
        cls, violations: list[SQLBaseError], ignore_mask: list[NoQaDirective]
    ) -> list[SQLBaseError]:
        """Returns whether to ignore error for line-range directives.

        The "ignore" list is assumed to ONLY contain NoQaDirectives where
        action is "enable" or "disable".
        """
        result = []
        for v in violations:
            # Find the directives that affect the violated rule "v", either
            # because they specifically reference it or because they don't
            # specify a list of rules, thus affecting ALL rules.
            ignore_rule = sorted(
                (
                    ignore
                    for ignore in ignore_mask
                    if not ignore.rules or (v.rule_code() in ignore.rules)
                ),
                key=lambda ignore: ignore.line_no,
            )
            # Determine whether to ignore the violation, based on the relevant
            # enable/disable directives.
            ignore, last_ignore = cls._should_ignore_violation_line_range(
                v.line_no, ignore_rule
            )
            if not ignore:
                result.append(v)
            # If there was a previous ignore which mean that we filtered out
            # a violation, then mark it as used.
            elif last_ignore:
                last_ignore.used = True

        return result

    def ignore_masked_violations(
        self, violations: list[SQLBaseError]
    ) -> list[SQLBaseError]:
        """Remove any violations specified by ignore_mask.

        This involves two steps:
        1. Filter out violations affected by single-line "noqa" directives.
        2. Filter out violations affected by disable/enable "noqa" directives.
        """
        ignore_specific = [ignore for ignore in self._ignore_list if not ignore.action]
        ignore_range = [ignore for ignore in self._ignore_list if ignore.action]
        violations = self._ignore_masked_violations_single_line(
            violations, ignore_specific
        )
        violations = self._ignore_masked_violations_line_range(violations, ignore_range)
        return violations

    def generate_warnings_for_unused(self) -> list[SQLBaseError]:
        """Generates warnings for any unused NoQaDirectives."""
        return [
            SQLUnusedNoQaWarning(
                line_no=ignore.line_no,
                line_pos=ignore.line_pos,
                description=f"Unused noqa: {ignore.raw_str!r}",
            )
            for ignore in self._ignore_list
            if not ignore.used
        ]



"""Individual segment parsers.

Matchable objects which return individual segments.
"""

from abc import abstractmethod
from collections.abc import Collection, Sequence
from typing import Any, Callable, Optional
from uuid import uuid4

import regex

from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.match_result import MatchResult
from sqlfluff.core.parser.matchable import Matchable
from sqlfluff.core.parser.segments import BaseSegment, RawSegment
from sqlfluff.core.parser.types import SimpleHintType


class BaseParser(Matchable):
    """An abstract class from which other Parsers should inherit."""

    # Meta segments are handled separately. All Parser elements
    # are assumed to be not meta.
    is_meta: bool = False

    @abstractmethod
    def __init__(
        self,
        raw_class: type[RawSegment],
        type: Optional[str] = None,
        optional: bool = False,
        # The following kwargs are passed on to the segment:
        trim_chars: Optional[tuple[str, ...]] = None,
        casefold: Optional[Callable[[str], str]] = None,
    ) -> None:
        self.raw_class = raw_class
        # Store instance_types rather than just type to allow
        # for multiple possible types to be supported in derivative
        # classes.
        self._instance_types: tuple[str, ...] = (type or raw_class.type,)
        self.optional = optional
        self._trim_chars = trim_chars
        self.casefold = casefold
        # Generate a cache key
        self._cache_key = uuid4().hex

    def cache_key(self) -> str:
        """Get the cache key for this parser.

        For parsers, they're unique per-instance.
        """
        return self._cache_key

    def is_optional(self) -> bool:
        """Return whether this element is optional."""
        return self.optional

    def segment_kwargs(self) -> dict[str, Any]:
        """Generates the segment_kwargs package for generating a matched segment."""
        segment_kwargs: dict[str, Any] = {}
        if self._instance_types:
            segment_kwargs["instance_types"] = self._instance_types
        if self._trim_chars:
            segment_kwargs["trim_chars"] = self._trim_chars
        if self.casefold:
            segment_kwargs["casefold"] = self.casefold
        return segment_kwargs

    def _match_at(self, idx: int) -> MatchResult:
        """Construct a MatchResult at a given index.

        This is a helper function for reuse by other parsers.
        """
        return MatchResult(
            matched_slice=slice(idx, idx + 1),
            matched_class=self.raw_class,
            segment_kwargs=self.segment_kwargs(),
        )


class TypedParser(BaseParser):
    """An object which matches and returns raw segments based on types."""

    def __init__(
        self,
        template: str,
        raw_class: type[RawSegment],
        type: Optional[str] = None,
        optional: bool = False,
        trim_chars: Optional[tuple[str, ...]] = None,
        casefold: Optional[Callable[[str], str]] = None,
    ) -> None:
        """Initialize a new instance of the class.

        Args:
            template (str): The template type.
            raw_class (type[RawSegment]): The raw segment class.
            type (Optional[str]): The type of the instance.
            optional (bool): Whether the instance is optional.
            trim_chars (Optional[tuple[str, ...]]): The characters to trim.
            casefold: (Optional[Callable[[str],str]]): The default casing used.

        Returns:
            None
        """
        # NB: the template in this case is the _target_ type.
        # The type kwarg is the eventual type.
        self.template = template
        # Pre-calculate the appropriate frozenset for matching later.
        self._target_types = frozenset((template,))
        super().__init__(
            raw_class=raw_class,
            optional=optional,
            trim_chars=trim_chars,
            casefold=casefold,
        )
        # NOTE: We override the instance types after initialising the base
        # class. We want to ensure that re-matching is possible by ensuring that
        # the `type` pre-matching is still present post-match even if it's not
        # part of the natural type hierarchy for the new `raw_class`.
        # The new `type` becomes the "primary" type, but the template will still
        # be part of the resulting `class_types`.
        # We do this here rather than in the base class to keep the dialect-facing
        # API the same.
        self._instance_types: tuple[str, ...] = ()
        # Primary type if set.
        if type is not None:
            self._instance_types += (type,)
        # New root types
        if type != raw_class.type:
            self._instance_types += (raw_class.type,)
        # Template type (if it's not in the subclasses of the raw_class).
        if not raw_class.class_is_type(template):
            self._instance_types += (template,)

    def __repr__(self) -> str:
        """Return a string representation of the TypedParser object."""
        return f"<TypedParser: {self.template!r}>"

    def simple(
        self, parse_context: ParseContext, crumbs: Optional[tuple[str, ...]] = None
    ) -> SimpleHintType:
        """Check if the matcher supports uppercase hash matching route.

        The TypedParser segment does not support matching against raw strings,
        but it does support matching against types. Matching is done against both the
        template and the resulting type, to support re-matching.

        Args:
            parse_context (ParseContext): The parse context.
            crumbs (Optional[tuple[str, ...]], optional): The crumbs.
            Defaults to None.

        Returns:
            SimpleHintType: A set of target types.
        """
        return frozenset(), self._target_types

    def match(
        self,
        segments: Sequence["BaseSegment"],
        idx: int,
        parse_context: "ParseContext",
    ) -> MatchResult:
        """Match against this matcher."""
        if segments[idx].is_type(self.template):
            return self._match_at(idx)
        return MatchResult.empty_at(idx)


class StringParser(BaseParser):
    """An object which matches and returns raw segments based on strings."""

    def __init__(
        self,
        template: str,
        raw_class: type[RawSegment],
        type: Optional[str] = None,
        optional: bool = False,
        trim_chars: Optional[tuple[str, ...]] = None,
        casefold: Optional[Callable[[str], str]] = None,
    ):
        self.template = template.upper()
        # Create list version upfront to avoid recreating it multiple times.
        self._simple = frozenset((self.template,))
        super().__init__(
            raw_class=raw_class,
            type=type,
            optional=optional,
            trim_chars=trim_chars,
            casefold=casefold,
        )

    def __repr__(self) -> str:
        return f"<StringParser: {self.template!r}>"

    def simple(
        self, parse_context: "ParseContext", crumbs: Optional[tuple[str, ...]] = None
    ) -> SimpleHintType:
        """Return simple options for this matcher.

        Because string matchers are not case sensitive we can
        just return the template here.
        """
        return self._simple, frozenset()

    def match(
        self,
        segments: Sequence["BaseSegment"],
        idx: int,
        parse_context: "ParseContext",
    ) -> MatchResult:
        """Match against this matcher.

        NOTE: We check that the segment is also code to avoid matching
        unexpected comments.
        """
        if segments[idx].raw_upper == self.template and segments[idx].is_code:
            return self._match_at(idx)
        return MatchResult.empty_at(idx)


class MultiStringParser(BaseParser):
    """An object which matches and returns raw segments on a collection of strings."""

    def __init__(
        self,
        templates: Collection[str],
        raw_class: type[RawSegment],
        type: Optional[str] = None,
        optional: bool = False,
        trim_chars: Optional[tuple[str, ...]] = None,
        casefold: Optional[Callable[[str], str]] = None,
    ):
        self.templates = {template.upper() for template in templates}
        # Create list version upfront to avoid recreating it multiple times.
        self._simple = frozenset(self.templates)
        super().__init__(
            raw_class=raw_class,
            type=type,
            optional=optional,
            trim_chars=trim_chars,
            casefold=casefold,
        )

    def __repr__(self) -> str:
        return f"<MultiStringParser: {self.templates!r}>"

    def simple(
        self, parse_context: "ParseContext", crumbs: Optional[tuple[str, ...]] = None
    ) -> SimpleHintType:
        """Return simple options for this matcher.

        Because string matchers are not case sensitive we can
        just return the templates here.
        """
        return self._simple, frozenset()

    def match(
        self,
        segments: Sequence["BaseSegment"],
        idx: int,
        parse_context: "ParseContext",
    ) -> MatchResult:
        """Match against this matcher.

        NOTE: We check that the segment is also code to avoid matching
        unexpected comments.
        """
        if segments[idx].is_code and segments[idx].raw_upper in self.templates:
            return self._match_at(idx)
        return MatchResult.empty_at(idx)


class RegexParser(BaseParser):
    """An object which matches and returns raw segments based on a regex."""

    def __init__(
        self,
        template: str,
        raw_class: type[RawSegment],
        type: Optional[str] = None,
        optional: bool = False,
        anti_template: Optional[str] = None,
        trim_chars: Optional[tuple[str, ...]] = None,
        casefold: Optional[Callable[[str], str]] = None,
        ignore_case: bool = True,
    ):
        # Store the optional anti-template
        self.template = template
        self.anti_template = anti_template
        self.ignore_case = ignore_case
        # Compile regexes upfront to avoid repeated overhead
        self._anti_template = regex.compile(
            anti_template or r"", regex.IGNORECASE if ignore_case else 0
        )
        self._template = regex.compile(template, regex.IGNORECASE if ignore_case else 0)
        super().__init__(
            raw_class=raw_class,
            type=type,
            optional=optional,
            trim_chars=trim_chars,
            casefold=casefold,
        )

    def __repr__(self) -> str:
        return f"<RegexParser: {self.template!r}>"

    def simple(
        self, parse_context: ParseContext, crumbs: Optional[tuple[str, ...]] = None
    ) -> None:
        """Does this matcher support a uppercase hash matching route?

        Regex segment does NOT for now. We might need to later for efficiency.
        """
        return None

    def match(
        self,
        segments: Sequence["BaseSegment"],
        idx: int,
        parse_context: "ParseContext",
    ) -> MatchResult:
        """Match against this matcher.

        NOTE: This method's case sensitivity is based on objects ignore_case property
        """
        _raw = segments[idx].raw_upper if self.ignore_case else segments[idx].raw
        result = self._template.match(_raw)
        if result:
            result_string = result.group(0)
            # Check that we've fully matched
            if result_string == _raw:
                # Check that the anti_template (if set) hasn't also matched
                if not self.anti_template or not self._anti_template.match(_raw):
                    return self._match_at(idx)
        return MatchResult.empty_at(idx)



"""Records of deprecated and removed config variables."""

import logging
from dataclasses import dataclass
from typing import Callable, Optional, Union

from sqlfluff.core.config.ini import coerce_value
from sqlfluff.core.errors import SQLFluffUserError
from sqlfluff.core.helpers.dict import (
    NestedStringDict,
    nested_dict_get,
    nested_dict_set,
    records_to_nested_dict,
)
from sqlfluff.core.types import ConfigMappingType, ConfigValueOrListType

# Instantiate the config logger
config_logger = logging.getLogger("sqlfluff.config")


def _translate_allow_implicit_indents(
    value: ConfigValueOrListType,
) -> ConfigValueOrListType:
    """Translate a deprecated ``allow_implicit_indents`` value.

    The old value was a boolean, where ``true`` maps to ``allow`` and
    ``false`` maps to ``forbid``. Ini config values are already coerced to
    real booleans before they reach here, but toml preserves the written
    type, so a quoted boolean (e.g. ``allow_implicit_indents = "false"``)
    arrives as a string. Coerce string values the same way ini does before
    testing them, so that a quoted ``"false"`` still maps to ``forbid``
    rather than being treated as a truthy string.

    Args:
        value: The deprecated config value.

    Returns:
        ``"allow"`` for truthy values, otherwise ``"forbid"``.
    """
    if isinstance(value, str):
        value = coerce_value(value)
    return "allow" if value else "forbid"


@dataclass
class _RemovedConfig:
    old_path: tuple[str, ...]
    warning: str
    new_path: Optional[tuple[str, ...]] = None
    translation_func: Optional[
        Callable[[ConfigValueOrListType], ConfigValueOrListType]
    ] = None

    @property
    def formatted_old_key(self) -> str:
        """Format the old key in a way similar to a config file."""
        return ":".join(self.old_path)

    @property
    def formatted_new_key(self) -> str:
        """Format the new key (assuming it exists) in a way similar to a config file."""
        assert self.new_path, (
            "`formatted_new_key` can only be called if a `new_path` is set."
        )
        return ":".join(self.new_path)


RemovedConfigMapType = dict[str, Union[_RemovedConfig, "RemovedConfigMapType"]]


REMOVED_CONFIGS = [
    _RemovedConfig(
        ("rules", "L003", "hanging_indents"),
        (
            "Hanging indents are no longer supported in SQLFluff "
            "from version 2.0.0 onwards. See "
            "https://docs.sqlfluff.com/en/stable/perma/hanging_indents.html"
        ),
    ),
    _RemovedConfig(
        ("rules", "max_line_length"),
        (
            "The max_line_length config has moved "
            "from sqlfluff:rules to the root sqlfluff level."
        ),
        ("max_line_length",),
        (lambda x: x),
    ),
    _RemovedConfig(
        ("rules", "tab_space_size"),
        (
            "The tab_space_size config has moved "
            "from sqlfluff:rules to sqlfluff:indentation."
        ),
        ("indentation", "tab_space_size"),
        (lambda x: x),
    ),
    _RemovedConfig(
        ("rules", "L002", "tab_space_size"),
        (
            "The tab_space_size config has moved "
            "from sqlfluff:rules to sqlfluff:indentation."
        ),
        ("indentation", "tab_space_size"),
        (lambda x: x),
    ),
    _RemovedConfig(
        ("rules", "L003", "tab_space_size"),
        (
            "The tab_space_size config has moved "
            "from sqlfluff:rules to sqlfluff:indentation."
        ),
        ("indentation", "tab_space_size"),
        (lambda x: x),
    ),
    _RemovedConfig(
        ("rules", "L004", "tab_space_size"),
        (
            "The tab_space_size config has moved "
            "from sqlfluff:rules to sqlfluff:indentation."
        ),
        ("indentation", "tab_space_size"),
        (lambda x: x),
    ),
    _RemovedConfig(
        ("rules", "L016", "tab_space_size"),
        (
            "The tab_space_size config has moved "
            "from sqlfluff:rules to sqlfluff:indentation."
        ),
        ("indentation", "tab_space_size"),
        (lambda x: x),
    ),
    _RemovedConfig(
        ("rules", "indent_unit"),
        (
            "The indent_unit config has moved "
            "from sqlfluff:rules to sqlfluff:indentation."
        ),
        ("indentation", "indent_unit"),
        (lambda x: x),
    ),
    _RemovedConfig(
        ("rules", "LT03", "operator_new_lines"),
        (
            "Use the line_position config in the appropriate "
            "sqlfluff:layout section (e.g. sqlfluff:layout:type"
            ":binary_operator)."
        ),
        ("layout", "type", "binary_operator", "line_position"),
        (lambda x: "trailing" if x == "before" else "leading"),
    ),
    _RemovedConfig(
        ("rules", "comma_style"),
        (
            "Use the line_position config in the appropriate "
            "sqlfluff:layout section (e.g. sqlfluff:layout:type"
            ":comma)."
        ),
        ("layout", "type", "comma", "line_position"),
        (lambda x: x),
    ),
    # LT04 used to have a more specific version of the same /config itself.
    _RemovedConfig(
        ("rules", "LT04", "comma_style"),
        (
            "Use the line_position config in the appropriate "
            "sqlfluff:layout section (e.g. sqlfluff:layout:type"
            ":comma)."
        ),
        ("layout", "type", "comma", "line_position"),
        (lambda x: x),
    ),
    _RemovedConfig(
        ("rules", "L003", "lint_templated_tokens"),
        "No longer used.",
    ),
    _RemovedConfig(
        ("core", "recurse"),
        "Removed as unused in production and unnecessary for debugging.",
    ),
    _RemovedConfig(
        ("rules", "references.quoting", "force_enable"),
        "No longer used. The dialects which used to block this rule, no longer do.",
    ),
    _RemovedConfig(
        ("indentation", "allow_implicit_indents"),
        (
            "The allow_implicit_indents config has been replaced by implicit_indents. "
            "Use 'forbid' (was false), 'allow' (was true), or 'require' instead."
        ),
        ("indentation", "implicit_indents"),
        _translate_allow_implicit_indents,
    ),
]

# Actually make a dict which matches the structure.
REMOVED_CONFIG_MAP = records_to_nested_dict(
    (removed_config.old_path, removed_config) for removed_config in REMOVED_CONFIGS
)


def validate_config_dict_for_removed(
    config: ConfigMappingType,
    logging_reference: str,
    removed_config: NestedStringDict[_RemovedConfig] = REMOVED_CONFIG_MAP,
    root_config_ref: Optional[ConfigMappingType] = None,
) -> None:
    """Validates a config dict against removed values.

    Where a value can be updated or translated, it mutates the config object.

    In general the `removed_config` & `root_config_ref` arguments are present
    only to enable recursion and shouldn't be necessary for general use of this
    function.
    """
    # If no root ref provided, then assume it's the config provided.
    # NOTE: During recursion, this should be set explicitly.
    root_config_ref = root_config_ref or config

    # Iterate through a copy of the config keys, so we can safely mutate
    # the underlying dict.
    for key in list(config.keys()):
        # Is there a removed config to compare to?
        if key not in removed_config:
            continue
        removed_value = removed_config[key]

        # If it's a section, recurse
        if isinstance(removed_value, dict):
            config_section = config[key]
            assert isinstance(config_section, dict), (
                f"Expected `{key}` to be a section not a value."
            )
            validate_config_dict_for_removed(
                config_section,
                logging_reference=logging_reference,
                removed_config=removed_value,
                root_config_ref=root_config_ref,
            )
            # If that validation resulted in an empty dict, also remove
            # the reference in this layer.
            if not config_section:
                del config[key]
            continue

        # Otherwise handle it directly.
        assert isinstance(removed_value, _RemovedConfig)

        # If there isn't a mapping option, just raise an error
        if not (removed_value.translation_func and removed_value.new_path):
            raise SQLFluffUserError(
                f"Config file {logging_reference!r} set an outdated config "
                f"value {removed_value.formatted_old_key}."
                f"\n\n{removed_value.warning}\n\n"
                "See https://docs.sqlfluff.com/en/stable/perma/"
                "configuration.html for more details."
            )

        # Otherwise perform the translation.
        # First check whether we have already set the new path?
        try:
            # Try and fetch a value at the new path.
            # NOTE: We don't actually handle the return value.
            nested_dict_get(root_config_ref, removed_value.new_path)
            # Raise an warning.
            config_logger.warning(
                f"\nWARNING: Config file {logging_reference} set a deprecated "
                f"config value `{removed_value.formatted_old_key}` (which can be "
                "migrated) but ALSO set the value it would be migrated to. The new "
                f"value (`{removed_value.formatted_new_key}`) takes precedence. "
                "Please update your configuration to remove this warning. "
                f"\n\n{removed_value.warning}\n\n"
                "See https://docs.sqlfluff.com/en/stable/perma/"
                "configuration.html for more details.\n"
            )
            # Remove the corresponding value from the dict object as invalid.
            del config[key]
            continue
        except KeyError:
            pass

        # If we haven't already set the new path then mutate and warn.
        old_value = config[key]
        assert not isinstance(old_value, dict), (
            f"Expected `{key}` to be a value not a section."
        )
        new_value = removed_value.translation_func(old_value)
        # NOTE: At the stage of emitting this warning, we may not yet
        # have set up red logging because we haven't yet loaded the config
        # file. For that reason, this error message has a bit more padding.
        config_logger.warning(
            f"\nWARNING: Config file {logging_reference} set a deprecated config "
            f"value `{removed_value.formatted_old_key}`. This will be "
            "removed in a later release. This has been mapped to "
            f"`{removed_value.formatted_new_key}` set to a value of "
            f"{new_value!r} for this run. "
            "Please update your configuration to remove this warning. "
            f"\n\n{removed_value.warning}\n\n"
            "See https://docs.sqlfluff.com/en/stable/perma/"
            "configuration.html for more details.\n"
        )
        # Write the new value and delete the old
        nested_dict_set(root_config_ref, removed_value.new_path, new_value)
        del config[key]



"""Tools for more complex analysis of SELECT statements."""

import logging
from collections.abc import Iterator
from dataclasses import dataclass, field
from enum import Enum
from functools import cached_property
from typing import Generic, NamedTuple, Optional, TypeVar, Union, cast

from sqlfluff.core.dialects.base import Dialect
from sqlfluff.core.dialects.common import (
    AliasInfo,
    get_from_clause_aliases,
    is_qualified,
)
from sqlfluff.core.parser import BaseSegment
from sqlfluff.dialects.dialect_ansi import ObjectReferenceSegment
from sqlfluff.utils.analysis.select import (
    SelectStatementColumnsAndTables,
    get_select_statement_info,
)
from sqlfluff.utils.functional import Segments, sp

analysis_logger = logging.getLogger("sqlfluff.rules.analysis")


# Segment types which directly are or contain selectables.
SELECTABLE_TYPES = (
    "with_compound_statement",
    "set_expression",
    "select_statement",
)

# Segment types which are likely to contain a subselect.
SUBSELECT_TYPES = (
    "merge_statement",
    "update_statement",
    "delete_statement",
    # NOTE: Values clauses won't have sub selects, but it's
    # also harmless to look, and they may appear in similar
    # locations. We include them here because they come through
    # the same code paths - although are likely to return nothing.
    "values_clause",
)


class QueryType(Enum):
    """Query type: Simple is just a query; WithCompound has CTE(s)."""

    Simple = 1
    WithCompound = 2


class WildcardInfo(NamedTuple):
    """Structure returned by Selectable.get_wildcard_info()."""

    segment: BaseSegment
    tables: list[str]


@dataclass
class Selectable:
    """A "SELECT" query segment."""

    selectable: BaseSegment
    dialect: Dialect

    def as_str(self) -> str:
        """String representation for logging/testing."""
        return self.selectable.raw

    @cached_property
    def select_info(self) -> Optional[SelectStatementColumnsAndTables]:
        """Returns SelectStatementColumnsAndTables on the SELECT."""
        if self.selectable.is_type("select_statement"):
            return get_select_statement_info(
                self.selectable, self.dialect, early_exit=False
            )
        else:  # DML or values_clause
            # This is a bit dodgy, but a very useful abstraction. Here, we
            # interpret a DML or values_clause segment as if it were a SELECT.
            # Someday, we may need to tweak this, e.g. perhaps add a separate
            # QueryType for this (depending on the needs of the rules that use
            # it.
            #
            # For more info on the syntax and behavior of VALUES and its
            # similarity to a SELECT statement with literal values (no table
            # source), see the "Examples" section of the Postgres docs page:
            # (https://www.postgresql.org/docs/8.2/sql-values.html).
            values = Segments(self.selectable)
            alias_expressions = values.children(sp.is_type("alias_expression"))
            table_aliases: list[AliasInfo] = []
            if alias_expressions:
                # A VALUES clause carries its alias as a direct child.
                for alias_expression in alias_expressions:
                    name = (
                        Segments(alias_expression)
                        .children()
                        .first(sp.is_type("naked_identifier", "quoted_identifier"))
                    )
                    table_aliases.append(
                        AliasInfo(
                            name[0].raw if name else "",
                            name[0] if name else None,
                            bool(name),
                            self.selectable,
                            alias_expression,
                            None,
                        )
                    )
            elif values.children(sp.is_type("from_expression")):
                # In some dialects (e.g. MySQL) the target table(s) of an
                # UPDATE are parsed as from_expression(s), so their aliases live
                # inside from_expression_elements (and join clauses, for
                # multi-table updates) rather than as direct children. Reuse the
                # canonical from-clause alias resolution instead of re-deriving
                # aliases here, so joined and nested targets resolve exactly as
                # they do in a SELECT.
                # https://github.com/sqlfluff/sqlfluff/issues/6147
                dialect_name = self.dialect.name if self.dialect else None
                table_aliases = [
                    alias
                    for _, alias in get_from_clause_aliases(
                        self.selectable, dialect_name
                    )
                ]

            if not table_aliases:
                # Preserve the longstanding behavior of always registering
                # one (possibly anonymous) alias for the DML target.
                table_aliases = [
                    AliasInfo("", None, False, self.selectable, None, None)
                ]

            return SelectStatementColumnsAndTables(
                select_statement=self.selectable,
                table_aliases=table_aliases,
                standalone_aliases=[],
                reference_buffer=[],
                select_targets=[],
                col_aliases=[],
                using_cols=[],
                table_reference_buffer=[],
            )

    def get_wildcard_info(self) -> list[WildcardInfo]:
        """Find wildcard (*) targets in the SELECT."""
        buff: list[WildcardInfo] = []
        # Some select-like statements don't have select_info
        # (e.g. test_exasol_invalid_foreign_key_from)
        if not self.select_info:  # pragma: no cover
            # TODO: Review whether to remove this.
            # Restructure of Exasol dialect means it no longer applies.
            return buff
        for seg in self.select_info.select_targets:
            if seg.get_child("wildcard_expression"):
                if "." in seg.raw:
                    # The wildcard specifies a target table.
                    table = seg.raw.rsplit(".", 1)[0]
                    buff.append(WildcardInfo(seg, [table]))
                else:
                    # The wildcard is unqualified (i.e. does not specify a
                    # table). This means to include all columns from all the
                    # tables in the query.
                    buff.append(
                        WildcardInfo(
                            seg,
                            [
                                (
                                    alias_info.ref_str
                                    if alias_info.aliased
                                    else alias_info.from_expression_element.raw
                                )
                                for alias_info in self.select_info.table_aliases
                                if alias_info.ref_str
                            ],
                        )
                    )
        return buff

    def find_alias(self, table: str) -> Optional[AliasInfo]:
        """Find corresponding table_aliases entry (if any) matching "table"."""
        alias_info = [
            t
            for t in (self.select_info.table_aliases if self.select_info else [])
            if t.aliased and t.ref_str == table
        ]
        assert len(alias_info) <= 1
        return alias_info[0] if alias_info else None


T = TypeVar("T", bound="Query")


@dataclass
class Query(Generic[T]):
    """A main SELECT query plus possible CTEs."""

    query_type: QueryType
    dialect: Dialect
    selectables: list[Selectable] = field(default_factory=list)
    ctes: dict[str, T] = field(default_factory=dict)
    # Parent scope. This query can "see" CTEs defined by parents.
    parent: Optional[T] = field(default=None)
    # subqueries are subselects in either the SELECT or FROM clause.
    subqueries: list[T] = field(default_factory=list)
    cte_definition_segment: Optional[BaseSegment] = field(default=None)
    cte_name_segment: Optional[BaseSegment] = field(default=None)
    is_subquery: Optional[bool] = None

    def __post_init__(self) -> None:
        # Once instantiated, set the `parent` attribute of any
        # subqueries and ctes. Some might already be set - but
        # we'll reset them anyway here.
        for subquery in self.subqueries:
            subquery.parent = self
            # We set this here to prevent a potential recursion error in RF03.
            subquery.is_subquery = True
        # NOTE: In normal operation, CTEs are typically set after
        # instantiation, and so for this method there aren't normally
        # any present. It is included here for completeness but not
        # covered in the test suite.
        # See `.from_segment()` for the way `parent` is set for CTEs.
        for cte in self.ctes.values():  # pragma: no cover
            cte.parent = self

    @property
    def children(self: T) -> list[T]:
        """Children could be CTEs, subselects or Others."""
        return list(self.ctes.values()) + self.subqueries

    def as_dict(self: T) -> dict:
        """Dict representation for logging/testing."""
        result: dict[str, Union[str, list[str], dict, list[dict]]] = {}
        if self.query_type != QueryType.Simple:
            result["query_type"] = self.query_type.name
        if self.selectables:
            result["selectables"] = [s.as_str() for s in self.selectables]
        if self.ctes:
            result["ctes"] = {k: v.as_dict() for k, v in self.ctes.items()}
        if self.subqueries:
            result["subqueries"] = [q.as_dict() for q in self.subqueries]
        return result

    def lookup_cte(self: T, name: str, pop: bool = True) -> Optional[T]:
        """Look up a CTE by name, in the current or any parent scope."""
        cte = self.ctes.get(name.upper())
        if cte:
            if pop:
                del self.ctes[name.upper()]
            return cte
        if self.parent:
            return self.parent.lookup_cte(name, pop)
        else:
            return None

    def crawl_sources(
        self: T,
        segment: BaseSegment,
        recurse_into: bool = True,
        pop: bool = False,
        lookup_cte: bool = True,
    ) -> Iterator[Union[str, T]]:
        """Find SELECTs, table refs, or value table function calls in segment.

        For each SELECT, yield a list of Query objects. As we find table
        references or function call strings, yield those.
        """
        found_nested_select = False
        for seg in segment.recursive_crawl(
            "table_reference",
            "set_expression",
            "select_statement",
            "values_clause",
            recurse_into=False,
            allow_self=False,
        ):
            # Crawl efficiently, don't recurse here. We do that later.
            # What do we have?
            # 1. If it's a table reference, work out whether it's to a CTE
            #    or to an external table.
            if seg.is_type("table_reference"):
                _seg = cast(ObjectReferenceSegment, seg)
                if not is_qualified(_seg, self.dialect.name) and lookup_cte:
                    cte = self.lookup_cte(_seg.raw, pop=pop)
                    if cte:
                        # It's a CTE.
                        yield cte
                # It's an external table reference.
                yield _seg.raw
            # 2. If it's some kind of more complex expression which is still
            #    valid in this position, generate an appropriate sub-select.
            else:
                assert seg.is_type(
                    "set_expression", "select_statement", "values_clause"
                )
                found_nested_select = True
                # Generate a subquery, referencing the current query
                # as the parent.
                yield self.__class__.from_segment(seg, self.dialect, parent=self)
        if not found_nested_select:
            # If we reach here, the SELECT may be querying from a value table
            # function, e.g. UNNEST(). For our purposes, this is basically the
            # same as an external table. Return the "table" part as a string.
            table_expr = segment.get_child("table_expression")
            if table_expr:
                yield table_expr.raw

    @classmethod
    def _extract_subqueries(
        cls: type[T], selectable: Selectable, dialect: Dialect
    ) -> Iterator[T]:
        """Given a Selectable, extract subqueries."""
        assert selectable.selectable.is_type(
            *SELECTABLE_TYPES,
            *SUBSELECT_TYPES,
        ), f"Found unexpected {selectable.selectable}"

        # For MERGE, UPDATE & DELETE, we should expect to find a sub select.
        for subselect in selectable.selectable.recursive_crawl(
            *SELECTABLE_TYPES,
            recurse_into=False,
            allow_self=False,
        ):
            # NOTE: We don't need to set the parent here, because it will
            # be set when attached to the parent later.
            yield cls.from_segment(subselect, dialect=dialect)

    @classmethod
    def from_root(cls: type[T], root_segment: BaseSegment, dialect: Dialect) -> T:
        """Given a root segment, find the first appropriate selectable and analyse."""
        selectable_segment = next(
            # Could be a Selectable or a MERGE
            root_segment.recursive_crawl(*SELECTABLE_TYPES, "merge_statement"),
            None,
        )
        assert selectable_segment, f"No selectable found in {root_segment.raw!r}."
        return cls.from_segment(selectable_segment, dialect=dialect)

    @classmethod
    def from_segment(
        cls: type[T],
        segment: BaseSegment,
        dialect: Dialect,
        parent: Optional[T] = None,
    ) -> T:
        """Recursively generate a query from an appropriate segment."""
        assert segment.is_type(*SELECTABLE_TYPES, *SUBSELECT_TYPES), (
            f"Invalid segment for `from_segment`: {segment}"
        )

        selectables = []
        subqueries = []
        cte_defs: list[BaseSegment] = []
        query_type = QueryType.Simple

        if segment.is_type("select_statement", *SUBSELECT_TYPES):
            # It's a select. Instantiate a Query.
            selectables = [Selectable(segment, dialect=dialect)]
        elif segment.is_type("set_expression"):
            # It's a set expression. There may be multiple selectables.
            for _seg in segment.recursive_crawl("select_statement", recurse_into=False):
                selectables.append(Selectable(_seg, dialect=dialect))
        else:
            # Otherwise it's a WITH statement.
            assert segment.is_type("with_compound_statement")
            query_type = QueryType.WithCompound
            for _seg in segment.recursive_crawl(
                # NOTE: We don't _specify_ set expressions here, because
                # all set expressions are made of selects, and we
                # want to look straight through to those child
                # expressions.
                "select_statement",
                recurse_into=False,
                no_recursive_seg_type="common_table_expression",
            ):
                selectables.append(Selectable(_seg, dialect=dialect))

            # We also need to handle CTEs
            for _seg in segment.recursive_crawl(
                "common_table_expression",
                recurse_into=False,
                # Don't recurse into any other WITH statements.
                no_recursive_seg_type="with_compound_statement",
            ):
                # Just store the segments for now.
                cte_defs.append(_seg)

        # Extract subqueries from any selectables.
        for selectable in selectables:
            # NOTE: If any VALUES clauses are present, they pass through here
            # safely without Exception. They won't yield any subqueries.
            subqueries += list(cls._extract_subqueries(selectable, dialect))

        # Instantiate the query
        outer_query = cls(
            query_type,
            dialect,
            selectables,
            parent=parent,
            subqueries=subqueries,
        )

        # If we don't have any CTEs, we can stop now.
        if not cte_defs:
            return outer_query

        # Otherwise build up the CTE map.
        ctes = {}
        for cte in cte_defs:
            # NOTE: This feels a little risky to just assume the first segment
            # is the name, but it's the same functionality we've run with for
            # a while.
            name_seg = cte.segments[0]
            name = name_seg.raw_normalized(False).upper()
            # Get the query out of it, just stop on the first one we find.
            try:
                inner_qry = next(
                    cte.recursive_crawl(
                        *SELECTABLE_TYPES,
                        "values_clause",
                        # Very rarely, we might find things like update
                        # clauses in here, handle them accordingly.
                        *SUBSELECT_TYPES,
                    ),
                )
            # If this fails it's because we didn't find anything "selectable"
            # in the CTE. Flag this up, but then carry on. It's likely something
            # strange (w.g. a Clickhouse WITH clause setting a with).
            except StopIteration:  # pragma: no cover
                # Log it as an issue, but otherwise skip this one.
                analysis_logger.info(f"Skipping unexpected CTE structure: {cte.raw!r}")
                continue
            qry = cls.from_segment(inner_qry, dialect=dialect, parent=outer_query)
            assert qry
            # Populate the CTE specific args.
            qry.cte_definition_segment = cte
            qry.cte_name_segment = name_seg
            # File it in the dictionary.
            ctes[name] = qry

        # Set the CTEs attribute on the outer.
        # NOTE: Because we're setting this after instantiation, it's important
        # that we've already set the `parent` value of the cte queries.
        outer_query.ctes = ctes
        return outer_query



"""Conditional Grammar."""

from collections.abc import Sequence
from typing import Union

from sqlfluff.core.parser.context import ParseContext
from sqlfluff.core.parser.grammar.base import BaseGrammar
from sqlfluff.core.parser.match_result import MatchResult
from sqlfluff.core.parser.segments import BaseSegment, Indent


class Conditional(BaseGrammar):
    """A grammar which is conditional on the parse context.

    | NOTE: The Conditional grammar is assumed to be operating
    | within a Sequence grammar, and some of the functionality
    | may not function within a different context.

    Args:
        *args: A meta segment which is instantiated
            conditionally upon the rules set.
        config_type: The area of the config that is used
            when evaluating the status of the given rules.
        rules: A set of `rule=boolean` pairs, which are
            evaluated when understanding whether conditions
            are met for this grammar to be enabled.

    Example:
        .. code-block:: python

            Conditional(Dedent, config_type="indent", indented_joins=False)

        This effectively says that if `indented_joins` in the "indent" section
        of the current config is set to `True`, then this grammar will allow
        a `Dedent` segment to be matched here. If `indented_joins` is set to
        `False`, it will be as though there was no `Dedent` in this sequence.

    | NOTE: While the Conditional grammar is set up to allow different
    | sources of configuration, it relies on configuration keys being
    | available within the ParseContext. Practically speaking only the
    | "indentation" keys are currently set up.
    """

    def __init__(
        self,
        meta: type[Indent],
        config_type: str = "indentation",
        **rules: Union[str, bool],
    ):
        """Initialize a new instance of the class.

        This method initializes an instance of the class with the provided
        arguments.

        Args:
            meta (type[Indent]): The meta argument.
            config_type (str, optional): The config_type argument. Defaults to
                "indentation".
            **rules (Union[str, bool]): The rules argument.
        """
        assert issubclass(meta, Indent), (
            "Conditional is only designed to work with Indent/Dedent segments."
        )
        self._meta = meta
        if not config_type:  # pragma: no cover
            raise ValueError("Conditional config_type must be set.")
        elif config_type not in ("indentation"):  # pragma: no cover
            raise ValueError(
                "Only 'indentation' is supported as a Conditional config_type."
            )
        if not rules:  # pragma: no cover
            raise ValueError("Conditional requires rules to be set.")
        self._config_type = config_type
        self._config_rules = rules
        super().__init__()

    def is_enabled(self, parse_context: ParseContext) -> bool:
        """Evaluate conditionals and return whether enabled."""
        # NOTE: Because only "indentation" is the only current config_type
        # supported, this code is much simpler that would be required in
        # future if multiple options are available.
        if self._config_type != "indentation":  # pragma: no cover
            raise ValueError(
                "Only 'indentation' is supported as a Conditional config_type."
            )
        config_section = parse_context.indentation_config
        # If any rules fail, return no match.
        for rule, val in self._config_rules.items():
            # Assume False if not set.
            conf_val = config_section.get(rule, False)
            # Coerce to boolean.
            if val != bool(conf_val):
                return False
        return True

    def match(
        self,
        segments: Sequence["BaseSegment"],
        idx: int,
        parse_context: "ParseContext",
    ) -> MatchResult:
        """If enabled, return a single insert of the new segment."""
        if not self.is_enabled(parse_context):
            return MatchResult.empty_at(idx)

        return MatchResult(
            matched_slice=slice(idx, idx), insert_segments=((idx, self._meta),)
        )



"""Defines the specification to implement a plugin."""

from abc import abstractmethod
from typing import TYPE_CHECKING, Any

import pluggy

from sqlfluff.core.plugin import plugin_base_name

if TYPE_CHECKING:  # pragma: no cover
    # NOTE: This import is against the normal import rules, but is here for strict
    # type checking. We have an exception for this in the import linter.
    from sqlfluff.core.rules.base import BaseRule

hookspec = pluggy.HookspecMarker(plugin_base_name)


class PluginSpec:
    """Defines the method signatures for plugin implementations."""

    @hookspec
    @abstractmethod
    def get_rules(self) -> list[type["BaseRule"]]:
        """Get plugin rules."""

    @hookspec
    @abstractmethod
    def load_default_config(self) -> dict[str, Any]:
        """Loads the default configuration for the plugin."""

    @hookspec
    @abstractmethod
    # TODO: This type annotation could probably be more specific but that would
    # require making the config info object something more like a namedTuple rather
    # than a dict.
    def get_configs_info(self) -> dict[str, dict[str, Any]]:
        """Get rule config validations and descriptions."""
